from omega.symbolic import temporal
# from dd import cudd


aut = temporal.Automaton()
aut.declare_variables(x='bool', y='bool')
u = aut.add_expr('x /\ y')
aut._clear_invalid_cache()
aut._cache_expr('x /\ y')
aut._fetch_expr(u)
aut._fetch_expr(u)
aut._add_expr('x')
del u
# cudd.reorder(aut.bdd)
aut._add_expr('y')
aut._add_expr('x \/ ~ y')
print(aut._bdd_to_expr)
aut._clear_invalid_cache()
print(aut._bdd_to_expr)



"""Syntactic check for GR(1) fragment."""
# Copyright 2015 by California Institute of Technology
# All rights reserved. Licensed under BSD-3.
#
from __future__ import absolute_import

import networkx as nx

from omega.logic import lexyacc
from omega.logic.ast import Nodes
from omega.logic import syntax as stx
from omega.symbolic import symbolic
from omega.logic import transformation as tx


parser = lexyacc.Parser()


def ltl_to_automaton(f):
    """Return `Automaton` from LTL formula `f` as `str`.

    Formula `f` must be in the form:

      A -+-> G

    where each of A, G is a conjunction of terms: `B`, `[]C`, `[]<>B`.
    For more details on `B, C`, see `split_gr1`.

    @type f: `str`
    @rtype: `symbolic.Automaton`
    """
    t = parser.parse(f)
    assert t.operator == '-+->'
    env, sys = t.operands
    d = {'assume': split_gr1(env),
         'assert': split_gr1(sys)}
    a = symbolic.Automaton()
    a.init['env'] = d['assume']['init']
    a.init['sys'] = d['assert']['init']
    a.action['env'] = d['assume']['[]']
    a.action['sys'] = d['assert']['[]']
    a.win['env'] = d['assume']['[]<>']
    a.win['sys'] = d['assert']['[]<>']
    return a


def split_gr1(s):
    """Return `dict` of GR(1) subformulas.

    The formula `f` is assumed to be a conjunction of expressions
    of the form:
      `B`, `[]C` or `[]<>B`
    where:
      - `C` can contain "next"
      - `B` cannot contain "next"

    @param f: temporal logic formula
    @type f: `str` or AST

    @return: conjunctions of formulas A, B as `str`,
    grouped by keys:
        `'init', '[]', '[]<>'`
    @rtype: `dict` of `str`: `list` of `str`
    """
    u = parser.parse(s)
    init = _split_gr1(u, 'init')
    action = _split_gr1(u, 'action')
    gf = list()
    _split_gr1(u, 'win', gf=gf)
    win = [v.flatten() for v in gf]
    d = dict(init=[init.flatten()],
             G=[action.flatten()],
             GF=win)
    _assert_admissible_operators(d)
    return d


def _split_gr1(u, context, gf=None):
    # terminal ?
    if hasattr(u, 'value'):
        if context == 'init':
            return u
        else:
            return Nodes.Bool('TRUE')
    # operator
    if u.operator == '/\\':
        v, w = u.operands
        p = _split_gr1(v, context, gf)
        q = _split_gr1(w, context, gf)
        return Nodes.Binary('/\\', p, q)
    assert u.operator != '/\\', u.operator
    if context == 'init':
        if u.operator == '[]':
            return Nodes.Bool('TRUE')
        else:
            return u
    if u.operator != '[]':
        return Nodes.Bool('TRUE')
    assert u.operator == '[]'
    (v,) = u.operands
    # terminal ?
    if hasattr(v, 'value'):
        if context == 'action':
            return v
        else:
            return Nodes.Bool('TRUE')
    assert hasattr(v, 'operator'), v
    if context == 'action':
        # GF ?
        if v.operator == '<>':
            return Nodes.Bool('TRUE')
        return v
    elif context == 'win':
        if v.operator != '<>':
            return
        assert v.operator == '<>'
        (w,) = v.operands
        gf.append(w)


def _assert_admissible_operators(d):
    action_ops = {'[]', '<>', 'U', 'V', 'R'}
    init_ops = set(action_ops)
    init_ops.add('X')
    operators = dict(init=init_ops, G=action_ops, GF=init_ops)
    for part, f in list(d.items()):
        ops = operators[part]
        for s in f:
            op = has_operator(s, ops)
            if op is None:
                continue
            raise AssertionError((
                'found inadmissible operator "{op}" '
                'in "{p}" formula. Parts:\n {d}').format(
                    op=op, p=part, d=d))


def has_operator(s, operators):
    u = parser.parse(s)
    return _has_operator(u, operators)


def _has_operator(u, operators):
    # terminal ?
    if hasattr(u, 'value'):
        return None
    # operator
    assert hasattr(u, 'operator')
    if u.operator in operators:
        return u.operator
    for v in u.operands:
        op = _has_operator(v, operators)
        if op is not None:
            return op
    return None


def split_gr1_old(f):
    """Earlier implementation of `split_gr1`.

    Does not preserve formula structure.
    Instead, it arranges conjunction operators as a
    binary tree.
    """
    # TODO: preprocess by applying syntactic identities: [][] = [] etc
    if stx.isinstance_str(f):
        t = f
    else:
        t = parser.parse(f)
    g = tx.Tree.from_recursive_ast(t)
    # collect boundary of conjunction operators
    Q = [g.root]
    b = list()  # use lists to preserve as much given syntactic order
    while Q:
        u = Q.pop()
        # terminal ?
        if not g.succ.get(u):
            b.append(u)
            continue
        # operator
        if u.operator == '/\\':
            # use `u.operands` instead of `g.successors`
            # to preserve original order
            Q.extend(u.operands)
        else:
            b.append(u)
    d = dict(init=list(), G=list(), GF=list())
    for u in b:
        # terminal ?
        if not g.succ.get(u):
            d['init'].append(u)
            continue
        # some operator
        if u.operator != '[]':
            d['init'].append(u)
            continue
        # G
        (v,) = u.operands
        # terminal in G ?
        if not g.succ.get(v):
            d['[]'].append(v)
            continue
        # some operator in G
        if v.operator == '<>':
            (w,) = v.operands
            d['[]<>'].append(w)
        else:
            # not a GF
            d['[]'].append(v)
    # assert only admissible temporal operators
    ops = {'[]', '<>', 'U', 'V', 'R'}
    operators = dict(G=ops)
    ops = set(ops)
    ops.add('X')
    operators.update(init=ops, GF=ops)
    for part, f in d.items():
        ops = operators[part]
        for u in f:
            op = has_operator(u, g, ops)
            if op is None:
                continue
            raise AssertionError((
                'found inadmissible operator "{op}" '
                'in "{f}" formula. Parts:\n {d}').format(
                    op=op, f=part, d=d))
    # conjoin (except for progress)
    init = conj(u.flatten() for u in reversed(d['init']))
    d['init'] = [init]
    safe = conj(u.flatten() for u in reversed(d['[]']))
    d['[]'] = [safe]
    # flatten individual progress formulas
    d['[]<>'] = [u.flatten() for u in d['[]<>']]
    return d


def conj(x):
    return ' /\ '.join(x)


def has_operator_old(u, g, operators):
    """Return `True` if AST `u` contains any `operators`."""
    try:
        if u.operator in operators:
            return u.operator
    except AttributeError:
        pass
    for v in nx.descendants(g, u):
        # terminal
        if not g.succ.get(v):
            continue
        # operator
        # is it temporal except for 'X' ?
        if v.operator in operators:
            return v.operator
    return None



"""Semi-symbolic automata."""
# Copyright 2015 by California Institute of Technology
# All rights reserved. Licensed under BSD-3.
#
from __future__ import absolute_import
import os
import logging
from pprint import pformat
import networkx as nx


logger = logging.getLogger(__name__)


class _SystemGraph(nx.MultiDiGraph):
    """Rooted multi-digraph with consistency methods.

    The attribute `initial_nodes` must contain nodes.
    """

    def __init__(self):
        super(_SystemGraph, self).__init__()
        self.initial_nodes = set()

    def assert_consistent(self):
        """Return `True` if attributes are conformant."""
        assert self.initial_nodes.issubset(self)

    def make_consistent(self):
        """Remove from attributes elements that are not nodes.

        Call this to update attributes like `initial_nodes`
        after removing nodes from `self`.
        """
        self.initial_nodes = self.initial_nodes.intersection(self)

    def dump(self, filename='out.pdf', rankdir='LR', prog='dot'):
        """Write to file using GraphViz via `pydot`.

        @param filename: path with exrension,
            for example: `"out.pdf"`.
        @param rankdir, prog: see `pydot.Dot.write`
        """
        name, ext = os.path.splitext(filename)
        pd = self._to_pydot()
        pd.set_rankdir(rankdir)
        pd.set_splines('true')
        pd.write(filename, format=ext[1:], prog=prog)

    def _to_pydot(self):
        return nx.drawing.nx_pydot.to_pydot(self)


class Automaton(_SystemGraph):
    """Alternating acceptor of (in)finite trees.

    An acceptor represents an indicator function of a set.
    The set may contain trees (unary trees are words).

    From all the possible paths in the acceptor's graph,
    some correspond to elements of the set it describes,
    but others do not belong to this set.

    In order to distinguish elements that can be generated by
    the graph, but are not contained in the set,
    two attributes are used:

      - path quantification
      - acceptance condition


    Attributes
    ==========

    Besides `initial_nodes` an [`Automaton`] has the attributes:

      - `"universal_nodes"` is a subset of nodes.
        Remaining nodes are existentially quantified.


      - `"acceptance"` that can be:

        - `"finite"`, signifying finite words or trees
        - `"Buchi"` (or `"weak Buchi"`)
        - `"coBuchi"`
        - `"generalized Buchi"`
        - `"Rabin"`
        - `"Streett"`
        - `"Muller"`
        - `"parity"`


      - `"accepting_sets"` that contains the sets of nodes defining
        the acceptance condition:

        - `set` of nodes for (weak/co-) Buchi and finite
          For co-Buchi this is the avoidance set (FG!)
        - `list` of `dict`s of `set`s of nodes for Rabin and Streett
          The keys should be: `"GF"` and `"FG!"`,
        - `list` of `set`s of nodes for Muller and generalized Buchi,
        - a 2-`tuple` of (min, max) colors for parity
          The coloring function is defined by directly annotating the nodes.

        For convenience, the above are initialized so that
        the automaton represent an empty language.
        The user is responsible for adhering to the above conventions.

        To check compliance to the above, call [`Automaton.check_sanity`].

        Deprecated: Note that by using `SubSet(automaton)` for each set,
        you can ensure consistency if more nodes are added later.
        Alternatively, write a function that checks consistency of
        each acceptance condition wrt the automaton's nodes.


      - `vars`: `dict` mapping symbols to domains.
        Each edge that has an existential node as source is
        labeled with either:
          - an valuation of `vars`, or
          - a formula over `vars`
        These define a state predicate at the target node.
        If no variables have universal quantification,
        then the automaton recognizes words.


      - `guards` defines the representation of edge labels and can be:

        - `"formula"` meaning that each edge is annotated with
          a Boolean formula as `str` or AST. (Typically a conjunction.)
        - `"enumeration"` meaning that each edge is annotated with a letter,
          as `set` or `dict` (closed world semantics).


    Remark
    ======

    In `ltl2dstar` documentation L denotes a "good" set.
    To avoid ambiguity, `dict`s with explicit modalities are used as labels.


    References
    ==========

    - Orna Kupferman and Moshe Vardi,
      "Safraless decision procedures", FOCS'05, pp.531--542
    - Def. 10.53, p.801, U{[BK08]
      <http://tulip-control.sourceforge.net/doc/bibliography.html#bk08>}
    - U{ltl2dstar<http://ltl2dstar.de/>} documentation


    See also
    ========

    `TransitionSystem`, `Transducer`
    """

    def __init__(self, acceptance='Buchi', dvars=None,
                 universal_nodes=None, guards='bool'):
        super(Automaton, self).__init__()
        if universal_nodes is None:
            universal_nodes = set()
        # init attributes
        self.universal_nodes = universal_nodes
        self.acceptance = acceptance
        self.accepting_sets = self._init_accepting_sets(acceptance)
        self.vars = dvars
        # explicit is better than implicit
        self.guards = guards

    def __str__(self):
        show_node_data = (self.acceptance == 'parity')
        is_unary = (not self.directions or len(self.directions) <= 1)

        def f(x): return pformat(x, indent=3)
        s = (
            '{hl}\n Alternating {self.acceptance} {word_tree} automaton\n'
            '{hl}\n'
            'Alphabet:\n'
            '{self.alphabet}\n\n'
            'Directions:\n'
            '{self.directions}\n\n'
            'Nodes:\n'
            '{nodes}\n\n'
            'Initial nodes:\n'
            '{init_nodes}\n\n'
            'Universal nodes:\n'
            '{universal_nodes}\n\n'
            'Existential nodes: the rest\n\n'
            'Accepting sets:\n'
            '{accepting_sets}\n\n'
            'Edges with guards:\n'
            '{edges}\n{hl}\n').format(
                hl=40 * '-',
                self=self,
                word_tree='word' if is_unary else 'tree',
                nodes=f(self.nodes(data=show_node_data)),
                init_nodes=f(self.initial_nodes),
                universal_nodes=f(self.universal_nodes),
                accepting_sets=f(self.accepting_sets),
                edges=f(self.edges(data=True)))
        return s

    def _to_pydot(self):
        # TODO: initial nodes
        g = nx.MultiDiGraph()
        for u, d in self.nodes(data=True):
            if u in self.universal_nodes:
                shape = 'box'
            else:
                shape = 'circle'
            if self.acceptance in {'finite', 'Buchi', 'coBuchi'}:
                peripheries = 2
            else:
                peripheries = 1
            g.add_node(u, shape=shape, peripheries=peripheries)
        for u, v, d in self.edges(data=True):
            label = ', '.join(
                '{k} = {v}'.format(k=k, v=v)
                for k, v in d.items()
                if k in self.directions or k in self.alphabet)
            g.add_edge(u, v, label=label)
        return nx.drawing.nx_pydot.to_pydot(g)

    def _init_accepting_sets(self, acceptance):
        """Return a `set`, `list` or other, depending on acceptance type."""
        if acceptance in {'finite', 'Buchi', 'coBuchi', 'weak Buchi'}:
            a = set()
        elif acceptance in {'Muller', 'generalized Buchi', 'Rabin', 'Streett'}:
            a = list()
        elif acceptance == 'parity':
            a = (None, None)
        else:
            raise ValueError('unknown acceptance: {s}'.format(s=acceptance))
        return a

    def assert_consistent(self):
        """Return `True` if conformant to conventions."""
        super(Automaton, self).asert_consistent()
        a = self.acceptance
        s = self.accepting_sets

        def f(x): return all(u in self for u in x)
        if a in {'finite', 'Buchi', 'coBuchi', 'weak Buchi'}:
            assert f(s)
        elif a in {'Muller', 'generalized Buchi'}:
            for x in s:
                assert f(s)
        elif a in {'Rabin', 'Streett'}:
            for x in s:
                assert len(x) == 2
                assert set(x) == {'[]<>', '<>[]!'}
                assert f(x['[]<>'])
                assert f(x['<>[]!'])
        elif a == 'parity':
            assert len(s) == 2
        else:
            raise Exception('Unknown acceptance: {a}'.format(a=a))
        for u, v, d in self.edges(data=True):
            t = self.alphabet
            r = self.directions
            if u in self.universal_nodes:
                t, r = r, t
            for k, v in d.items():
                if k in t:
                    _check_value(v, t[k])
                assert k not in r
        return True


class TransitionSystem(_SystemGraph):
    """Enumerated or semi-symbolic transition system.

    Attributes:

      - `owner`: who selects the next node
        `in ("env", "sys")`
      - `vars`: a `dict` that maps each variable to a domain
      - `env_vars`: universally quantified variables.


    Node and edge labels
    ====================

    A node or edge attribute can be:
      - an assignment of a value to a variable, or
      - a formula as `str` (key: `"formula"`)

    These are over:
      - unprimed `vars` for nodes
      - primed and unprimed `vars` for edges

    The variable name is a key from `vars` (possibly primed).
    An assignment is interpreted as the corresponding conjunction.
    So a partial assignment is understood using open world semantics.
    """

    def __init__(self):
        super(TransitionSystem, self).__init__()
        self.owner = 'sys'
        self.vars = dict()
        self.env_vars = set()

    def __str__(self):
        return (
            '{hl}\n'
            'Transition system:\n'
            '{hl}\n'
            'Owner:\n'
            '{self.owner}\n'
            'Variables:\n'
            '{self.vars}\n'
            'Environment variables:\n'
            '{self.env_vars}\n'
            'Nodes with assignments to vars:\n'
            '{nodes}\n'
            'Initial nodes:\n'
            '{self.initial_nodes}\n'
            'Edges with actions:\n'
            '{edges}\n'
            '{hl}\n').format(
                hl=40 * '-',
                self=self,
                nodes=_dumps_nodes(self),
                edges=pformat(self.edges(data=True), indent=3))

    def _to_pydot(self):
        g = nx.MultiDiGraph()
        for u, d in self.nodes(data=True):
            label = ', '.join(
                '{k} = {v}'.format(k=k, v=v)
                for k, v in d.items()
                if k in self.vars)
            label = '"{u}\n{label}"'.format(u=u, label=label)
            g.add_node(u, label=label, shape='box', **d)
        for u, v, d in self.edges(data=True):
            f = d.get('formula')
            if f is None:
                label = '""'
            else:
                label = '"{f}"'.format(f=f)
            g.add_edge(u, v, label=label, **d)
        return nx.drawing.nx_pydot.to_pydot(g)

    def assert_consistent(self):
        """Return `True` if attributes are conformant."""
        # TODO: check type consistency of formulas
        super(TransitionSystem, self).assert_consistent()
        assert self.owner in {'env', 'sys'}
        assert set(self.env_vars).issubset(self.vars)
        for u, d in self.nodes(data=True):
            for k, v in d.items():
                if k in self.vars:
                    _check_value(v, self.vars[k])
        for u, v, d in self.edges(data=True):
            for k, v in d.items():
                # primed ?
                if k.endswith("'"):
                    var = k[:-1]
                else:
                    var = k
                if var in self.vars:
                    _check_value(v, self.vars[var])
        return True


def _dumps_nodes(g):
    """Return string of graph `g` nodes."""
    r = list()
    for u, d in g.nodes(data=True):
        s = '\t Node: {u}, {values}\n'.format(
            u=u,
            values=', '.join(
                '{k} = {v}'.format(k=k, v=v)
                for k, v in d.items()))
        r.append(s)
    return ''.join(r)


def _check_value(v, dom):
    """Assert that value `v` is in domain `dom`."""
    if dom == 'bool':
        assert v in (0, 1, False, True)
    elif isinstance(dom, tuple):
        assert dom[0] <= v <= dom[1]
    else:
        raise TypeError('unknown domain "{dom}"'.format(dom=dom))



#!/usr/bin/env python
import logging
from dd import bdd as _bdd
import networkx as nx
from omega.logic import bitvector as bv
from omega.symbolic import bdd as bddizer
from nose import tools as nt


logger = logging.getLogger(__name__)
logging.getLogger('tulip.ltl_parser_log').setLevel(logging.WARNING)
h = logging.StreamHandler()
log = logging.getLogger('omega.logic.bitvector')
log.setLevel(logging.WARNING)
log.addHandler(h)
logging.getLogger('astutils').setLevel('ERROR')

# TODO: test truncation, test ALU up to 32 bits


def test_bitvector_name_conflicts():
    # no conflict
    t = {'x_1': dict(type='bool', owner='env'),
         'x': dict(type='int', dom=(0, 1), owner='sys',
                   signed=False, width=1)}
    bv._add_bitnames(t)
    assert 'bitnames' in t['x'], t
    assert t['x']['bitnames'] == ['x_0'], t
    # integer `x` mapped to `x_0`, conflict with Boolean `x_0`
    t['x_0'] = t.pop('x_1')
    with nt.assert_raises(AssertionError):
        bv._add_bitnames(t)


def test_bitfield_to_int_states():
    t = {'x': dict(type='bool'),
         'y': dict(type='int', signed=True, bitnames=['y0', 'y1', 'y2'])}
    g = nx.DiGraph()
    w = [dict(x=1, y0=1, y1=1, y2=0),
         dict(x=0, y0=0, y1=1, y2=1),
         dict(x=1, y0=1, y1=1, y2=1)]
    for i, d in enumerate(w):
        g.add_node(i, state=d)
    g.add_edges_from([(0, 1), (1, 2), (2, 1)])
    h = bv.bitfield_to_int_states(g, t)
    assert set(h) == {0, 1, 2}
    assert set(h.edges()) == {(0, 1), (1, 2), (2, 1)}
    z = [dict(x=1, y=3), dict(x=0, y=-2), dict(x=1, y=-1)]
    for u, d in h.nodes(data=True):
        assert z[u] == d['state']


def test_bitfields_to_ints():
    t = {'x': dict(type='bool'),
         'y': dict(type='int', signed=True, bitnames=['y0', 'y1', 'y2'])}
    bit_state = {'x': 0, 'y0': 0, 'y1': 1, 'y2': 1}
    d = bv.bitfields_to_ints(bit_state, t)
    assert d == {'x': 0, 'y': -2}


t = {'a': {'type': 'int',
           'signed': True,
           'bitnames': ['a0', 'a1']},
     'b': {'type': 'int',
           'signed': True,
           'bitnames': ['b0', 'b1']},
     'x': {'type': 'int',
           'signed': True,
           'bitnames': ['x@0.0.3', 'x@1']},
     'y': {'type': 'int',
           'signed': True,
           'bitnames': ['y0', 'y1', 'y2']},
     'q': {'type': 'bool'},
     'r': {'type': 'bool'}}
parser = bv._parser


def test_flatten_comparator():
    """Test arithmetic predicates."""
    # tree = parser.parse('x + 1 <= 2')
    # print(repr(tree) + '\n')
    log.setLevel(logging.ERROR)
    # (in)equality
    assert parser.parse('a = 1').flatten(t=t) == '$ 1 ! | ^ a0 1 | ^ a1 0 0'
    assert parser.parse('a != 1').flatten(t=t) == '$ 1 | ^ a0 1 | ^ a1 0 0'
    # '<' comparator
    f = parser.parse('a < 1').flatten(t=t)
    # print(f)
    assert (f == ('$ 7 '
                  '^ ^ a0 ! 1 1 '
                  '| & a0 ! 1 & ^ a0 ! 1 1 '
                  '^ ^ a1 ! 0 ? 1 '
                  '| & a1 ! 0 & ^ a1 ! 0 ? 1 '
                  '^ ^ a1 ! 0 ? 3 '
                  '| & a1 ! 0 & ^ a1 ! 0 ? 3 '
                  '^ ! ^ a1 0 ? 5')), f
    # TODO: more bits, larger numbers


def test_flatten_arithmetic():
    """Test arithmetic functions."""
    # addition
    mem = list()
    res = parser.parse('a + 1').flatten(t=t, mem=mem)
    assert res == ['? 0', '? 2', '? 4'], res
    assert mem == ['^ ^ a0 1 0',
                   '| & a0 1 & ^ a0 1 0',
                   '^ ^ a1 0 ? 1',
                   '| & a1 0 & ^ a1 0 ? 1',
                   '^ ^ a1 0 ? 3',
                   '| & a1 0 & ^ a1 0 ? 3'], mem
    # TODO: subtraction


def test_flatten_quantifiers():
    # single qvar
    s = '\A a: True'
    r = parser.parse(s).flatten(t=t)
    assert r == ' \A & a0 a1 1', r
    s = '\E y: False'
    r = parser.parse(s).flatten(t=t)
    assert r == ' \E & & y0 y1 y2 0', r
    s = '\E y: x = y'
    r = parser.parse(s).flatten(t=t)
    eq = parser.parse('x = y').flatten(t=t)
    r_ = ' \E & & y0 y1 y2 {eq}'.format(eq=eq)
    assert r == r_, (r, r_)
    # multiple qvars
    s = '\E a, b: True'
    r = parser.parse(s).flatten(t=t)
    assert r == ' \E & & & a0 a1 b0 b1 1', r
    s = '\E a, x: x - a > 0'
    r = parser.parse(s).flatten(t=t)
    h = parser.parse('x - a > 0').flatten(t=t)
    r_ = ' \E & & & a0 a1 x@0.0.3 x@1 {h}'.format(h=h)
    assert r == r_, (r, r_)
    s = '\A r: r | ! q'
    r = parser.parse(s).flatten(t=t)
    h = parser.parse('r | ! q').flatten(t=t)
    r_ = ' \A r {h}'.format(h=h)
    assert r == r_, (r, r_)


def test_flatten_substitution():
    # single pair
    s = '\S a / b: True'
    r = parser.parse(s).flatten(t=t)
    assert r == ' \S $4 a0 b0 a1 b1 1', r
    s = '\S q / r: False'
    r = parser.parse(s).flatten(t=t)
    assert r == ' \S $2 q r 0', r
    s = '\S b / x: x - y <= 0'
    r = parser.parse(s).flatten(t=t)
    h = parser.parse('x - y <= 0').flatten(t=t)
    r_ = ' \S $4 b0 x@0.0.3 b1 x@1 {h}'.format(h=h)
    assert r == r_, r
    # multiple pairs
    s = '\S a / b,  q / r:  False'
    r = parser.parse(s).flatten(t=t)
    r_ = ' \S $6 a0 b0 a1 b1 q r 0'
    assert r == r_, r
    # swap: order of substitutions should be preserved
    s = '\S q / r, a / b:  False'
    r = parser.parse(s).flatten(t=t)
    r_ = ' \S $6 q r a0 b0 a1 b1 0'
    assert r == r_, r
    # more complex expr
    s = '\S a / b, q / r:  r | ! (a != b)'
    r = parser.parse(s).flatten(t=t)
    h = parser.parse('r | ! (a != b)').flatten(t=t)
    r_ = ' \S $6 a0 b0 a1 b1 q r {h}'.format(h=h)
    assert r == r_, r
    # invalid input
    c = [
        '\S a / q: True',
        '\S q / a: True',
        '\S a / y: True']
    for s in c:
        print(s)
        with nt.assert_raises(AssertionError):
            u = parser.parse(s)
            u.flatten(t=t)


def test_div_mul_expr():
    mem = list()
    res = parser.parse('4 / 2').flatten(t=t, mem=mem)
    r = _evaluate_result(res, mem)
    assert r == 2, r
    mem = list()
    res = parser.parse('-8 / 3').flatten(t=t, mem=mem)
    r = _evaluate_result(res, mem)
    assert r == -2, r
    mem = list()
    res = parser.parse('4 * 2').flatten(t=t, mem=mem)
    r = _evaluate_result(res, mem)
    assert r == 8, r
    mem = list()
    res = parser.parse('4 % 3').flatten(t=t, mem=mem)
    r = _evaluate_result(res, mem)
    assert r == 1, r
    mem = list()
    res = parser.parse('4 % -3').flatten(t=t, mem=mem)
    r = _evaluate_result(res, mem)
    assert r == 1, r


def test_division():
    # 1 / 1 = 1 * 1 + 0
    x = ['1', '0']
    y = ['1', '0']
    _divide(x, y, q=1, r=0)
    # -2 / 1 = -2 * 1 + 0
    x = ['0', '1']
    y = ['1', '0']
    _divide(x, y, q=-2, r=0)
    # 0 / 1 == 0 * 1 + 0
    x = ['0', '0']
    y = ['1', '0']
    _divide(x, y, q=0, r=0)
    # 2 / -2 = -1 * (-2) + 0
    x = ['0', '1', '0']
    y = ['0', '1', '1']
    _divide(x, y, q=-1, r=0)
    #
    # negative numbers
    # -4 / 2 = -2 * 2 + 0
    x = ['0', '0', '1']
    y = ['0', '1', '0']
    _divide(x, y, q=-2, r=0)
    # 4 / -2 = -2 * (-2) + 0
    x = ['0', '0', '1', '0']
    y = ['0', '1']
    _divide(x, y, q=-2, r=0)
    #
    # C99 semantics
    # 4 / -3 = -1 * (-3) + 1
    x = ['0', '0', '1', '0']
    y = ['1', '0', '1']
    _divide(x, y, q=-1, r=1)
    # -4 / 3 = -1 * 3 - 1
    x = ['0', '0', '1', '1']
    y = ['1', '1', '0']
    _divide(x, y, q=-1, r=-1)
    #
    # divisor > dividend
    # 4 / 5 = 0 * 5 + 4
    x = ['0', '0', '1', '0']
    y = ['1', '0', '1', '0']
    _divide(x, y, q=0, r=4)
    # -4 / -5 = 0 * (-5) -4
    x = ['0', '0', '1', '1']
    y = ['1', '1', '0', '1']
    _divide(x, y, q=0, r=-4)


def _divide(x, y, q, r):
    quo, rem, mem = bv.restoring_divider(x, y)
    q_ = _evaluate_result(quo, mem)
    r_ = _evaluate_result(rem, mem)
    assert q_ == q, (q_, r_)
    assert r_ == r, (q_, r_)


def test_multiplier():
    # LSB ... MSB
    # 0 * 0 = 0
    x = ['0', '0']
    y = list(x)
    res, mem = bv.multiplier(x, y, start=0)
    r = _evaluate_result(res, mem)
    assert r == 0, r
    # 1 * 0 = 0
    x = ['0', '0']
    y = ['1', '0']
    res, mem = bv.multiplier(x, y, start=0)
    r = _evaluate_result(res, mem)
    assert r == 0, r
    # -1 * 1 = -1
    x = ['1', '1']
    y = ['1', '0']
    res, mem = bv.multiplier(x, y, start=0)
    r = _evaluate_result(res, mem)
    # 1 * -1 = -1
    x = ['1', '0']
    y = ['1', '1']
    res, mem = bv.multiplier(x, y, start=0)
    r = _evaluate_result(res, mem)
    assert r == -1, r
    # -2 * 2 = -4
    x = ['0', '1', '1']
    y = ['0', '1', '0']
    res, mem = bv.multiplier(x, y, start=0)
    r = _evaluate_result(res, mem)
    assert r == -4, r
    # 8 * 8 = 64
    x = ['0', '0', '0', '1', '0']
    y = list(x)
    res, mem = bv.multiplier(x, y, start=0)
    r = _evaluate_result(res, mem)
    assert r == 64, r
    # -16 * 5 = -80
    x = ['0', '0', '0', '0', '1', '1']
    y = ['1', '0', '1', '0']
    res, mem = bv.multiplier(x, y, start=0)
    r = _evaluate_result(res, mem)
    assert r == -80, r


def _evaluate_result(result, memory):
    """Return integer, given a result and memory w/o variables.

    @type result: `list`
    @type memory: `list`
    """
    bdd = _bdd.BDD({'x': 0})
    n = len(memory) + 1
    mem = ' '.join(memory)
    bits = list()
    for bit in result:
        s = '$ {n} {mem} {bit}'.format(n=n, mem=mem, bit=bit)
        u = bddizer.add_expr(s, bdd)
        bits.append(u)
    bits = [b == 1 for b in bits]
    # print([int(b) for b in bits])
    j = bv.twos_complement_to_int(bits)
    return j


def test_fixed_shift():
    # 0 is LSB
    x = ['0', '1', '0']
    c = 1
    r = bv.fixed_shift(x, c, left=True, logical=None)
    assert r == ['0', '0', '1'], r
    c = 2
    r = bv.fixed_shift(x, c, left=True, logical=None)
    assert r == ['0', '0', '0'], r
    c = 1
    r = bv.fixed_shift(x, c, left=False, logical=False)
    assert r == ['1', '0', '0'], r
    x = ['0', '1', '1']
    # unsigned
    r = bv.fixed_shift(x, c, left=False, logical=True)
    assert r == ['1', '1', '0']
    # signed
    r = bv.fixed_shift(x, c, left=False, logical=False)
    assert r == ['1', '1', '1']


def test_barrel_shifter():
    x = ['x0', 'x1', 'x2', 'x3']
    y = ['y0', 'y1']
    m = ['& ! y0 x0',
         '| & ! y0 x1 & y0 x0',
         '| & ! y0 x2 & y0 x1',
         '| & ! y0 x3 & y0 x2',
         '& ! y1 ? 0',
         '& ! y1 ? 1',
         '| & ! y1 ? 2 & y1 ? 0',
         '| & ! y1 ? 3 & y1 ? 1']
    # stage -1
    z, mem = bv.barrel_shifter(x, y, -1)
    assert z == x
    assert mem == list()
    # stage 0
    z, mem = bv.barrel_shifter(x, y, 0)
    assert z == ['? 0', '? 1', '? 2', '? 3']
    assert len(mem) == 4
    assert mem == m[:4]
    # stage 1
    z, mem = bv.barrel_shifter(x, y, 1)
    print(z)
    print(bv._format_mem(mem))
    assert z == ['? 4', '? 5', '? 6', '? 7']
    assert len(mem) == 8
    assert mem == m
    # fun: (2**n)-bit x
    # n = 5
    # x = ['x{i}'.format(i=i) for i in xrange(2**n)]
    # y = ['y{i}'.format(i=i) for i in xrange(n)]
    # z, mem = sl.barrel_shifter(x, y)
    # print(z)
    # print(sl._format_mem(mem))


def test_truncate():
    x = list('01010100011')
    n = 4
    r = bv.truncate(x, n)
    assert len(r) == n
    assert r == list('0101')


def test_twos_complement_for_var():
    t = dict(
        x=dict(type='bool'),
        y=dict(type='int', dom=(-2, 1), signed=True,
               bitnames=['y0', 'y1']),
        z=dict(type='int', dom=(0, 1), signed=False,
               bitnames=['z0']),
        w=dict(type='int', dom=(-1, 0), signed=True,
               bitnames=['w0', 'w1']),
        p=dict(type='int', dom=(-1, -1), signed=False,
               bitnames=['p0']),
        q=dict(type='int', dom=(-1, 0), signed=False,
               bitnames=['q0']))
    with nt.assert_raises(Exception):
        bv.var_to_twos_complement('r', t)
    with nt.assert_raises(TypeError):
        bv.var_to_twos_complement('x', t)
    y = bv.var_to_twos_complement('y', t)
    assert y == ['y0', 'y1'], y
    z = bv.var_to_twos_complement('z', t)
    assert z == ['z0', '0'], z
    w = bv.var_to_twos_complement('w', t)
    assert w == ['w0', 'w1'], w
    p = bv.var_to_twos_complement('p', t)
    assert p == ['p0', '1'], p
    with nt.assert_raises(AssertionError):
        bv.var_to_twos_complement('q', t)


def test_twos_complement_for_int():
    f = bv.twos_complement_to_int
    g = bv.int_to_twos_complement
    pairs = {0: ['0', '0'],
             1: ['1', '0'],
             2: ['0', '1', '0'],
             5: ['1', '0', '1', '0'],
             -1: ['1', '1'],
             -2: ['0', '1', '1'],
             -3: ['1', '0', '1']}
    for k, v in pairs.items():
        assert g(k) == v
        assert k == f(v)
    pairs = {-1: ['1', '1', '1'],
             -2: ['0', '1', '1', '1'],
             2: ['0', '1', '0', '0']}
    for k, v in pairs.items():
        assert k == f(v)


def test_abs():
    x = ['1', '0']
    r, mem = bv.abs_(x)
    a = _evaluate_result(r, mem)
    assert a == 1, a
    x = ['0', '1']
    r, mem = bv.abs_(x)
    a = _evaluate_result(r, mem)
    assert a == 2, a
    x = ['1', '1']
    r, mem = bv.abs_(x)
    a = _evaluate_result(r, mem)
    assert a == 1, a


def test_equalize_width():
    x = list('0101')
    y = list('10')
    p, q = bv.equalize_width(x, y)
    assert len(p) == len(q) == 4
    for a in p[len(x):]:
        assert a == x[-1]
    for a in q[len(y):]:
        assert a == y[-1]


def test_sign_extension():
    t = {'a': {'type': 'int',
               'signed': True,
               'bitnames': ['a0', 'a1']},
         'b': {'type': 'int',
               'signed': True,
               'bitnames': ['b0', 'b1']}}
    with nt.assert_raises(ValueError):
        bv.sign_extension(['1'], 2)
    with nt.assert_raises(ValueError):
        bv.sign_extension(['1', '1'], 0)
    assert bv.sign_extension(['1', '1'], 3) == ['1', '1', '1']
    assert bv.sign_extension(['1', '0'], 3) == ['1', '0', '0']
    with nt.assert_raises(AssertionError):
        mem = list()
        parser.parse('a + 1').flatten(t=t)
    with nt.assert_raises(AssertionError):
        parser.parse('b < -1').flatten(t=t, mem=mem)


def test_ite():
    # ite connective
    x = 'x'
    y = 'y'
    z = 'z'
    s = bv.ite_connective(x, y, z)
    r = '$ 2 x | & y ? 0 & z ! ? 0'
    assert s == r, s
    # ite connective flattening
    t = {'x': {'type': 'bool'},
         'y': {'type': 'bool'},
         'z': {'type': 'bool'}}
    p = parser.parse('ite( x, y, z)')
    s = p.flatten(t=t)
    assert s == r, s
    with nt.assert_raises(AssertionError):
        bv.ite_connective('a', ['b', 'c'], 'd')
    # ite function
    x = 'x'
    y = ['y0', 'y1']
    z = ['z0', 'z1']
    r, mem = bv.ite_function(x, y, z, start=0)
    correct = [
        'x',
        '| & y0 ? 0 & z0 ! ? 0',
        '| & y1 ? 0 & z1 ! ? 0']
    assert mem == correct, mem
    assert r == ['? 1', '? 2']
    # flip
    r, more_mem = bv.ite_function(x, z, y, start=len(mem))
    mem.extend(more_mem)
    correct.extend([
        'x',
        '| & z0 ? 3 & y0 ! ? 3',
        '| & z1 ? 3 & y1 ! ? 3'])
    assert mem == correct, mem
    assert r == ['? 4', '? 5']
    # ite function flattening
    t = {'x': {'type': 'bool'},
         'y': {'type': 'int',
               'signed': True,
               'bitnames': ['a0', 'a1']},
         'z': {'type': 'int',
               'signed': True,
               'bitnames': ['b0', 'b1']}}
    mem = list()
    r = p.flatten(t=t, mem=mem)
    assert mem == [
        'x',
        '| & a0 ? 0 & b0 ! ? 0',
        '| & a1 ? 0 & b1 ! ? 0']
    assert r == ['? 1', '? 2']
    # b, c of different length
    with nt.assert_raises(AssertionError):
        bv.ite_function('x', ['y0'], ['z0', 'z1', 'z2'], start=0)


def test_init_to_logic():
    # bool
    d = dict(type='bool', init='False')
    c = bv._init_to_logic('x', d)
    assert c == 'x <=> False', c
    # number
    d = dict(type='other', init=5)
    c = bv._init_to_logic('y', d)
    assert c == 'y = 5', c
    # TODO: create equivalent test
    # array
    # d = dict(dom='numerical', length=3, init=init)
    # c = logic.init_to_logic('x', d)
    # assert c == ['x0 = 5', 'x1 = 5', 'x2 = 5'], c


def test_bool_eq_number():
    t = dict(x=dict(type='bool', owner='env'))
    # wrong
    s = 'x <=> 0'
    tree = parser.parse(s)
    with nt.assert_raises(AssertionError):
        tree.flatten(t=t)
    # correct
    s = 'x <=> false'
    tree = parser.parse(s)
    r = tree.flatten(t=t)
    assert r == ' ! ^ x 0 ', r


def test_mixed_fol_bitblasted():
    t = dict(x=dict(type='bool', owner='sys'),
             y=dict(type='int', dom=(0, 3), owner='sys'))
    t = bv.bitblast_table(t)
    s = '(x /\ y_0) \/ (y < 0)'
    tree_0 = parser.parse(s)
    q = 'y < 0'
    tree_1 = parser.parse(q)
    f0 = tree_0.flatten(t=t)
    f1 = tree_1.flatten(t=t)
    assert f0 == ' |  & x y_0  {f1} '.format(f1=f1), (f0, f1)


def test_type_invariants():
    t = dict(x=dict(type='int', dom=(0, 3), init=2))
    t = bv.bitblast_table(t)
    init, action = bv.type_invariants(t)
    init_ = dict(x=['x = 2', '(0 <= x)  /\  (x <= 3)'])
    assert init == init_, init
    s = (
         "    (0 <= x)  /\  (x <= 3)"
         "  /\ (0 <= x') /\  (x' <= 3)")
    action_ = dict(x=[s])
    assert action == action_, action


def test_prefix_parser():
    parser = bddizer.Parser()
    nodes = parser.nodes
    # &, !
    e = '& x ! y'
    t = parser.parse(e)
    assert isinstance(t, nodes.Operator), t
    assert t.operator == '&', t.operator
    assert len(t.operands) == 2, t.operands
    a, b = t.operands
    assert isinstance(a, nodes.Var), a
    assert a.value == 'x', a.value
    assert isinstance(b, nodes.Operator), b
    assert b.operator == '!', b.operator
    assert len(b.operands) == 1, b.operands
    (y,) = b.operands
    assert isinstance(y, nodes.Var), y
    assert y.value == 'y', y
    # memory buffer: $
    e = '&   $ 1 ! x  y'
    t = parser.parse(e)
    # &
    assert isinstance(t, nodes.Operator)
    assert t.operator == '&', t.operator
    assert len(t.operands) == 2, t.operands
    a, b = t.operands
    assert isinstance(a, nodes.Buffer), a
    assert len(a.memory) == 1, a.memory
    # ! x
    (notx,) = a.memory
    assert isinstance(notx, nodes.Operator), notx
    assert len(notx.operands) == 1, notx.operands
    # x
    (x,) = notx.operands
    assert isinstance(x, nodes.Var), x
    assert x.value == 'x', x.value
    # y
    assert isinstance(b, nodes.Var), b
    assert b.value == 'y', b.value
    # register: ?
    e = '$ 2   x    & y  ? 0'
    t = parser.parse(e)
    assert isinstance(t, nodes.Buffer), t
    assert len(t.memory) == 2, t.memory
    x, c = t.memory
    # x
    assert isinstance(x, nodes.Var), x
    assert x.value == 'x', x.value
    # &
    assert isinstance(c, nodes.Operator), c
    assert c.operator == '&', c.operator
    assert len(c.operands) == 2, c.operands
    y, reg = c.operands
    assert isinstance(y, nodes.Var), y
    assert y.value == 'y', y.value
    assert isinstance(reg, nodes.Register), reg
    assert reg.value == '0', reg.value


def test_flatten_memory_nodes():
    # buffer
    nodes = bddizer.DebugNodes()
    x = nodes.Var('x')
    mem = [x, x, x]
    b = nodes.Buffer(mem)
    assert b.memory is mem, b.memory
    f = b.flatten()
    assert f == '\nbuffer[3](\n x,\n x,\n x)', f
    # register
    reg = nodes.Register('1')
    f = reg.flatten()
    assert f == 'reg[1]', f


def test_make_table():
    d = dict(x=(0, 2), y='bool', w='bool', z=(-2, 5))
    env_vars = {'x', 'y'}
    t = bv.make_table(d, env_vars)
    t_ = dict(
        x=dict(type='saturating', owner='env', dom=(0, 2)),
        y=dict(type='bool', owner='env', dom=None),
        z=dict(type='saturating', owner='sys', dom=(-2, 5)),
        w=dict(type='bool', owner='sys', dom=None))
    assert t == t_, (t, t_)


if __name__ == '__main__':
    test_type_invariants()



"""Translation of past to future LTL.


References
==========

Orna Lichtenstein, Amir Pnueli, Lenore Zuck
    "The glory of the past"
    Conf. on Logics of Programs
    LNCS Vol.193, pp.196--218, 1985

Zohar Manna, Amir Pnueli
    "The anchored version of the temporal framework"
    Linear time, branching time and partial order in
    logics and models of concurrency
    LNCS Vol.354, pp.201--284

Yonit Kesten, Amir Pnueli, Li-on Raviv
    "Algorithmic verification of linear temporal logic specifications"
    Int. Colloquium on Automata, Languages, and Programming (ICALP)
    LNCS Vol.1443, pp.1--16, 1998

Roderick Bloem, Barbara Jobstmann, Nir Piterman,
    Amir Pnueli, Yaniv Sa'ar
    "Synthesis of reactive(1) designs"
    Journal of Computer and System Sciences
    Vol.78, No.3, pp.911--938, 2012
"""
# Copyright 2015 by California Institute of Technology
# All rights reserved. Licensed under BSD-3.
#
import logging
from omega.logic.ast import Nodes as _Nodes
from omega.logic import lexyacc
from omega.logic.syntax import conj


logger = logging.getLogger('astutils').setLevel(logging.ERROR)


class Nodes(_Nodes):
    """LTL AST with flatteners to future LTL."""

    class Operator(_Nodes.Operator):
        def flatten(self, context=None, *arg, **kw):
            # ite both function and connective
            if self.operator == 'ite':
                x, y, z = self.operands
                a = x.flatten(context='bool', *arg, **kw)
                b = y.flatten(context=context, *arg, **kw)
                c = z.flatten(context=context, *arg, **kw)
                return ''.join([
                    self.operator,
                    '(',
                    ', '.join((a, b, c)),
                    ')'])
            return super(Nodes.Operator, self).flatten(
                context=context, *arg, **kw)

    class Unary(_Nodes.Unary):
        def flatten(self, *arg, **kw):
            # type checking
            c = kw.get('context')
            until = kw.get('until')
            # recurse
            op = self.operator
            if op in ('-X', '--X'):
                x = self.operands[0]
                return _flatten_previous(op, x, *arg, **kw)
            if op != 'X':
                assert c == 'bool', (c, op)
            if op == '-[]':
                x = Nodes.Bool('True')
                y = Nodes.Unary('~', self.operands[0])
                a = (x, y)
                r = _flatten_since(a, *arg, **kw)
                return '(~ {r})'.format(r=r)
            elif op == '-<>':
                a = (Nodes.Bool('True'), self.operands[0])
                return _flatten_since(a, *arg, **kw)
            elif op == '[]' and until:
                x = Nodes.Bool('True')
                y = Nodes.Unary('~', self.operands[0])
                a = (x, y)
                r = _flatten_until(a, *arg, **kw)
                return '(~ {r})'.format(r=r)
            elif op == '<>' and until:
                a = (Nodes.Bool('True'), self.operands[0])
                return _flatten_until(a, *arg, **kw)
            return super(Nodes.Unary, self).flatten(*arg, **kw)

    class Binary(_Nodes.Binary):
        def flatten(self, *arg, **kw):
            # type checking
            c = kw.get('context')
            assert c == 'bool', c
            until = kw.get('until')
            # recurse
            op = self.operator
            if op == 'S':
                return _flatten_since(
                    self.operands, *arg, **kw)
            elif op == 'U' and until:
                return _flatten_until(
                    self.operands, *arg, **kw)
            return super(Nodes.Binary, self).flatten(*arg, **kw)

    class Comparator(_Nodes.Comparator):
        def flatten(self, *arg, **kw):
            # type checking
            c = kw.get('context')
            assert c == 'bool', (c, str(self))
            # change context
            kw['context'] = 'arithmetic'
            # recurse
            return super(Nodes.Comparator, self).flatten(*arg, **kw)

    class Arithmetic(_Nodes.Arithmetic):
        def flatten(self, *arg, **kw):
            # type checking
            c = kw.get('context')
            assert c == 'arithmetic', c
            # change context
            kw['context'] = 'arithmetic'
            # recurse
            return super(Nodes.Arithmetic, self).flatten(*arg, **kw)

    class Var(_Nodes.Var):
        def flatten(self, testers=None, context=None,
                    previous=None, strong=None, free_init=None,
                    *arg, **kw):
            var = self.value
            if previous is None:
                return var
            # previous N
            assert testers is not None
            assert context == 'bool', (context, self.value)
            previous + 1  # isinstance(previous, int)  ?
            var_prev = '{name}_prev{i}'.format(
                name=var, i=previous)
            # add tester
            init, trans = _make_tester_for_previous(
                var_prev, var, context, strong)
            testers[var_prev] = dict(
                type='bool',  # previous applies only to bool vars
                init=init, trans=trans, win=None)
            return var_prev


def _flatten_previous(op, x, testers, context,
                      previous=0, strong=None, *arg, **kw):
    """Translate expression with "previous" as main operator."""
    # propagate ?
    # (if child is terminal)
    # added benefit: shares some history vars among subformulas
    strong = (op == '--X')
    propagate = (
        len(x) == 1)
    if propagate:
        previous += 1
        return x.flatten(testers=testers, context=context,
                         previous=previous, strong=strong, *arg, **kw)
    # create tester here
    assert context == 'bool', context
    assert len(x) > 1, 'operand is an operator'
    expr = x.flatten(testers=testers, context=context, *arg, **kw)
    # bottom-up counting is safe
    # `len` *must* be called after `flatten`
    # note that each `x_prev` occupies an "aux" index
    i = len(testers)
    var_prev = '_aux{i}'.format(i=i)
    # create tester
    init, trans = _make_tester_for_previous(
        var_prev, expr, context, strong)
    testers[var_prev] = dict(
        type='bool',
        init=init, trans=trans, win=None)
    return var_prev


def _make_tester_for_previous(var, expr, context, strong):
    """Return temporal tester for "previous"."""
    # select operator
    if context == 'bool':
        op = '<=>'
        # strong "previous" operator "--X" ?
        if strong:
            init = '(~ {var})'.format(var=var)
        else:
            init = var
    else:
        raise Exception(
            'unknown context: "{c}"'.format(c=context))
    # translate "previous" in future LTL
    trans = '((X {var}) {op} {expr})'.format(
        var=var, op=op, expr=expr)
    return init, trans


def _flatten_since(operands, testers, context, *arg, **kw):
    """Translate expression with "since" as main operator."""
    assert context == 'bool', context
    x, y = operands
    p = x.flatten(testers=testers, context=context, *arg, **kw)
    q = y.flatten(testers=testers, context=context, *arg, **kw)
    i = len(testers)
    var = '_aux{i}'.format(i=i)
    init = '({var} <=> {q})'.format(var=var, q=q)
    trans = (
        '('
        '(X {var}) <=> ('
        '    (X {q}) \/ ((X {p}) /\ {var})'
        '))').format(
            var=var, p=p, q=q)
    testers[var] = dict(
        type='bool',
        init=init, trans=trans, win=None)
    return var


def _flatten_until(operands, testers, context, *arg, **kw):
    assert context == 'bool', context
    x, y = operands
    p = x.flatten(testers=testers, context=context, *arg, **kw)
    q = y.flatten(testers=testers, context=context, *arg, **kw)
    i = len(testers)
    var = '_aux{i}'.format(i=i)
    trans = (
        '('
        '{var} <=> ('
        '    ({q}) \/ (({p}) /\ (X {var}))'
        '))').format(
            var=var, p=p, q=q)
    win = '(({q}) \/ ~ {var})'.format(var=var, q=q)
    testers[var] = dict(
        type='bool',
        init='True', trans=trans, win=win)
    return var


# LTL parser that translates past to future.
parser = lexyacc.Parser(nodes=Nodes)


def translate(s, debug=False, until=False):
    """Translate action formula `s` with past to future LTL.

    Return:
      - history and prophecy variable symbol table
      - translated formula
      - initial condition of temportal testers
      - conjunction `c` of translated formula with
        transition relations of temporal testers.
      - list of recurrence goals

    If formula `s` is an action (in the sense of TLA),
    then the returned formula `c` is also an action.

    Note that if two subformulas are identical,
    then a fresh variable will be used for each one.
    The only exception are variables, for example "-X p".

    @type s: `str`
    @param debug: ensures repeatable ordering
        of new subformulas, to enable testing.
    @param until: add prophecy variables for "until" too
    @return: `(dvars, translated, init, action)`
    @rtype: `tuple`
    """
    tree = parser.parse(s)
    testers = dict()
    context = 'bool'
    r = tree.flatten(testers=testers, context=context,
                     until=until)
    if debug:
        ci = sorted(d['init'] for d in testers.values())
        ct = sorted(d['trans'] for d in testers.values())
        win = [d['win'] for d in testers.values()
               if d['win'] is not None]
    else:
        ci = (d['init'] for d in testers.values())
        ct = (d['trans'] for d in testers.values())
        win = [d['win'] for d in testers.values()
               if d['win'] is not None]
    init = conj(ci)
    trans = conj(ct)
    # collect new vars
    dvars = dict()
    for var_prev, d in testers.items():
        dtype = d['type']
        dom = d.get('dom')
        dvars[var_prev] = dict(type=dtype, dom=dom, owner='sys')
    return dvars, r, init, trans, win


def map_translate(c, **kw):
    """Apply `translate` to all items of container `c`."""
    all_vars = dict()
    f = list()
    init = list()
    action = list()
    win = list()
    for s in c:
        dvars, r, i, a, w = translate(s, **kw)
        all_vars.update(dvars)
        f.append(r)
        init.append(i)
        action.append(a)
        win.extend(w)
    return all_vars, f, init, action, win



"""Realizability of properties expressed with the operator `WhilePlusHalf`.

The operator `WhilePlusHalf` implies the stepwise form:

```tla
[] \/ ~ Earlier(EnvNext(x, y, x', y'))
   \/ /\ Earlier(SysNext(x, y, x', y'))
      /\ \E u:  SysNext(x, y, u, y')
```

The controllable step operator in `omega.symbolic.fixpoint` has the form:

```tla
\E v:  \A u:
    /\ SysNext(x, y, u, v)
    /\ EnvNext(x, y, u, v) => Target(u, v)
```

for the choice of options `moore=True, plus_one=True`.
This script adapts the specification to the synthesizer,
by preprocessing the specified actions to derive appropriate actions to
pass to the synthesizer.
"""
# Copyright 2017 by California Institute of Technology
# All rights reserved. Licensed under 3-clause BSD.
#
from omega.games import gr1
from omega.symbolic import temporal as trl


aut = trl.Automaton()
aut.declare_variables(x=(1, 5), y=(1, 5))
aut.varlist['env'] = ['x']
aut.varlist['sys'] = ['y']
aut.prime_varlists()
# environment formulas
env_init = aut.to_bdd('x = 1')
env_next = aut.to_bdd('''
    \/ (x' \in 1..5)
    \/ (x' = x)
    ''')
env_live = aut.to_bdd(' x = 2 ')
# component formulas
sys_init = aut.to_bdd('y \in 1..5  /\  (y = x)')
sys_next = aut.to_bdd('''
    \/ (y' = x)
    \/ (x' = x  /\  y' = y)
    ''')
sys_live = aut.to_bdd(' y = 2 ')


def exist(*arg):
    *vrs, u = arg
    return aut.exist(vrs, u)

def forall(*arg):
    *vrs, u = arg
    return aut.forall(vrs, u)


sys_init_synth = exist('x', sys_init) | ~ exist('x', 'y', env_init)
sys_init_synth &= sys_init | ~ env_init

sys_next_synth = (
    exist("x'", sys_next) &
    forall("x'", sys_next | ~ env_next))

aut.moore = True  # a synthesis option that describes the implementation
aut.plus_one = True  # a specification formula option, affects
    # init form and
    # the steady state (win set via the controllable step operator)
aut.qinit = '\E \A'  # whether state is disjoint or initially shared:
    # a synthesis option that describes the kind of implementation

# initial conditions
aut.init['env'] = env_init
aut.init['sys'] = sys_init_synth
# actions
aut.action['env'] = env_next
aut.action['sys'] = sys_next_synth
# liveness requirements
aut.win['<>[]'] = [~ env_live]
aut.win['[]<>'] = [sys_live]


fx_iterates = gr1.solve_streett_game(aut)
gr1.make_streett_transducer(*fx_iterates, aut)



"""Functional synthesis.

Convert a predicate over input and output variables to multiple predicates
that depend only on input variables, one predicate for each output variable.

For example, given a BDD for the predicate `f(x1, x2, y1, y2)` where `x1, x2`
are regarded as inputs and `y1, y2` as outputs, create two BDDs for the
predicates `g(x1, x2)` and `h(x1, x2)` such that

```
\/ ~ /\ y1 = g(x1, x2)
     /\ y2 = h(x1, x2)
\/ f(x1, x2, y1, y2)
```


References
==========

Roderick Bloem, Barbara Jobstmann, Nir Piterman,
Amir Pnueli, Yaniv Sa'ar
    "Synthesis of reactive(1) designs"
    Journal of Computer and System Sciences
    Vol.78, No.3, pp.911--938, 2012
"""
# Copyright 2016-2017 by California Institute of Technology
# All rights reserved. Licensed under 3-clause BSD.
#
from dd import cudd as _bdd


def make_functions(r, vrs, bdd):
    """Extract functions for `vrs` from relation `r`.

    @param r: relation
    @param vrs: variables to extract functions for.
        These variables should be Boolean-valued, and
        declared in the manager `bdd`.
    @type bdd: `BDD`

    @return: `dict(function=g, care_set=care)` where
        `g` and `care` are BDD nodes.
    """
    supp = bdd.support(r)
    outputs = set(vrs)
    outputs &= supp
    functions = dict()
    for yp in set(outputs):
        outputs.remove(yp)
        g, care = extract_function(r, yp, outputs, bdd)
        sub = {yp: g}
        r = bdd.let(sub, r)
        functions[yp] = dict(function=g, care_set=care)
        # assert
        support = bdd.support(r)
        assert yp not in support, (yp, support)
        support = bdd.support(g)
        assert yp not in support, (yp, support)
    # no `vrs` in any support
    for d in functions.values():
        g = d['function']
        support = bdd.support(g)
        common = support.intersection(vrs)
        assert not common, (common, support)
    return functions


def extract_function(f, yp, outputs, bdd):
    """Extract function for variabe `yp`."""
    u = bdd.exist(outputs, f)
    p = bdd.let({yp: True}, u)
    n = bdd.let({yp: False}, u)
    # subtract overlap
    p = p & ~ n
    n = n & ~ p
    # optimize cofactors
    supp_p = bdd.support(p)
    supp_u = bdd.support(n)
    inputs = supp_p | supp_u
    for z in inputs:
        pz = bdd.exist(inputs, p)
        nz = bdd.exist(inputs, n)
        disjoint = ((pz & nz) == bdd.false)
        if disjoint:
            p, n = pz, nz
    # restrict
    care = (p & ~ n) | (n & ~ p)
    g = _bdd.restrict(p, care)
    # g = p
    return g, care


def collect_functions(functions, r=None):
    """Return functions for next values of latches.

    @param functions: `dict` as returned by `make_functions`
    """
    if r is None:
        r = dict()
    r.update(
        (var, d['function'])
        for var, d in functions.items())
    return r


def report_function_statistics(outputs, memory):
    """Print BDD sizes for output and memory BDDs."""
    care = next(outputs.values())['care_set']
    print('care set nodes: {n}'.format(n=len(care)))
    # collect all flip-flops
    functions = list()
    functions.extend(d['function'] for d in outputs.values())
    functions.extend(d['function'] for d in memory.values())
    n = _bdd.count_nodes(functions)
    print('total nodes (shared) in functions: {n}'.format(n=n))



"""Tests for gr1 fragment utilities."""
from nose.tools import assert_raises
from omega import gr1


def test_split_gr1():
    # init
    f = '(x > 0) /\ (y + 1 < 2)'
    d = gr1.split_gr1(f)
    assert d['init'] == ['( ( x > 0 ) /\ ( ( y + 1 ) < 2 ) )'], d
    assert d['G'] == ['( TRUE /\ TRUE )'], d
    assert d['GF'] == [], d
    # safety
    f = '[]((x > 0) /\ (z = 3 + y))'
    d = gr1.split_gr1(f)
    assert d['init'] == ['TRUE'], d
    assert d['G'] == ['( ( x > 0 ) /\ ( z = ( 3 + y ) ) )'], d
    assert d['GF'] == [], d
    # recurrence
    f = '[]<>(x > 0)'
    d = gr1.split_gr1(f)
    assert d['init'] == ['TRUE'], d
    assert d['G'] == ['TRUE'], d
    assert d['GF'] == ['( x > 0 )'], d
    # all together
    f = (
        '(x > 0) /\ (y + 1 < 2) /\ '
        '[]( (X y) > 0) /\ '
        '[]<>((z - x <= 0) \/ (p => q))')
    d = gr1.split_gr1(f)
    s = (
        '( ( ( ( x > 0 ) /\ ( ( y + 1 ) < 2 ) ) '
        '/\ TRUE ) /\ TRUE )')
    assert d['init'] == [s], d
    assert d['G'] == [
        '( ( ( TRUE /\ TRUE ) /\ ( ( X y ) > 0 ) ) /\ TRUE )'], d
    assert d['GF'] == [
        '( ( ( z - x ) <= 0 ) \/ ( p => q ) )'], d
    # not in fragment
    with assert_raises(AssertionError):
        gr1.split_gr1('[]( [] p )')
    with assert_raises(AssertionError):
        gr1.split_gr1('<>( [] p )')
    with assert_raises(AssertionError):
        gr1.split_gr1('(X p ) /\ ( [] p )')
    with assert_raises(AssertionError):
        gr1.split_gr1('[]<> ( x /\ (X y) )')


if __name__ == '__main__':
    test_split_gr1()



"""Tests for `omega.logic`."""
from nose.tools import assert_raises

from omega.logic import past
from omega.logic import syntax as stx


parser = past.parser


def test_flatten_previous_var():
    testers = dict()
    context = 'bool'
    x = past.Nodes.Var('x')
    s = past._flatten_previous(
        '--X', x, testers, context)
    assert s == 'x_prev1', s
    # tester
    assert 'x_prev1' in testers, testers
    d = testers['x_prev1']
    assert isinstance(d, dict), d
    # init
    a = d.get('init')
    b = '(~ x_prev1)'
    assert a == b, a
    # trans
    a = d.get('trans')
    b = '((X x_prev1) <=> x)'
    assert a == b, a


def test_flatten_previous_boolean():
    testers = dict()
    context = 'bool'
    x = past.Nodes.Var('x')
    y = past.Nodes.Var('y')
    previous_x = past.Nodes.Unary('--X', x)
    e = past.Nodes.Binary('/\\', previous_x, y)
    s = e.flatten(
        testers,
        context=context)
    assert s == '( x_prev1 /\ y )', s
    # tester
    assert 'x_prev1' in testers, testers
    d = testers['x_prev1']
    assert isinstance(d, dict), d
    # init
    a = d.get('init')
    b = '(~ x_prev1)'
    assert a == b, a
    # trans
    a = d.get('trans')
    b = '((X x_prev1) <=> x)'
    assert a == b, a


def test_flatten_previous_boolean_expr():
    testers = dict()
    context = 'bool'
    # "-X (x /\ y)"
    x = past.Nodes.Var('x')
    y = past.Nodes.Var('y')
    conj = past.Nodes.Binary('/\\', x, y)
    e = past.Nodes.Unary('--X', conj)
    s = e.flatten(
        testers=testers,
        context=context)
    assert s == '_aux0', s
    # tester
    assert '_aux0' in testers, testers
    d = testers['_aux0']
    assert isinstance(d, dict), d
    # init
    a = d.get('init')
    b = '(~ _aux0)'
    assert a == b, a
    # trans
    a = d.get('trans')
    b = '((X _aux0) <=> ( x /\ y ))'
    assert a == b, a


def test_flatten_nested_previous():
    testers = dict()
    context = 'bool'
    # "-X (q /\ -X (3 = (p + 3)))"
    p = past.Nodes.Var('p')
    q = past.Nodes.Var('q')
    n = past.Nodes.Num('3')
    plus = past.Nodes.Arithmetic('+', p, n)
    eq = past.Nodes.Comparator('=', n, plus)
    prev_eq = past.Nodes.Unary('--X', eq)
    conj = past.Nodes.Binary('/\\', q, prev_eq)
    e = past.Nodes.Unary('-X', conj)
    s = e.flatten(
        testers=testers,
        context=context)
    assert s == '_aux1', s
    #
    # tester for "-X (p + 3 = 3)" (_aux1)
    #
    assert '_aux1' in testers, testers
    d = testers['_aux1']
    assert isinstance(d, dict), d
    # init
    a = d.get('init')
    b = '_aux1'
    assert a == b, a
    # trans
    a = d.get('trans')
    b = '((X _aux1) <=> ( q /\ _aux0 ))'
    assert a == b, a
    #
    # tester for whole formula (_aux1)
    #
    assert '_aux1' in testers, testers
    d = testers['_aux1']
    assert isinstance(d, dict), d
    # init
    a = d.get('init')
    b = '_aux1'
    assert a == b, a
    # trans
    a = d.get('trans')
    b = '((X _aux1) <=> ( q /\ _aux0 ))'
    assert a == b, a


def test_flatten_since():
    testers = dict()
    context = 'bool'
    p = past.Nodes.Var('p')
    q = past.Nodes.Var('q')
    since = past.Nodes.Binary('S', p, q)
    s = since.flatten(
        testers=testers,
        context=context)
    assert s == '_aux0', s
    # tester
    assert '_aux0' in testers
    d = testers['_aux0']
    assert isinstance(d, dict), d
    # init
    a = d.get('init')
    b = '(_aux0 <=> q)'
    assert a == b, a
    # trans
    a = d.get('trans')
    b = '((X _aux0) <=> (    (X q) \/ ((X p) /\ _aux0)))'
    assert a == b, a


def test_past_parser_boolean():
    s = '-X a /\ b'
    dvars, r, init, trans, win = past.translate(s)
    init_ = 'a_prev1'
    assert init == init_, init
    r_ = '( a_prev1 /\ b )'
    assert r == r_, r
    trans_ = '((X a_prev1) <=> a)'
    assert trans == trans_, trans


def test_parser_strong_weak_previous():
    # strong previous
    tree = parser.parse('--X p')
    assert len(tree) == 2, tree
    assert hasattr(tree, 'operator'), tree
    op = tree.operator
    assert op == '--X', op
    assert hasattr(tree, 'operands'), tree
    c = tree.operands
    assert len(c) == 1, c
    (p,) = c
    assert hasattr(p, 'value'), p
    v = p.value
    assert v == 'p', v
    # weak previous
    tree = parser.parse('-X p')
    assert len(tree) == 2, tree
    assert hasattr(tree, 'operator'), tree
    op = tree.operator
    assert op == '-X', op
    assert hasattr(tree, 'operands'), tree
    c = tree.operands
    assert len(c) == 1, c
    (p,) = c
    assert hasattr(p, 'value'), p
    v = p.value
    assert v == 'p', v


def test_parser_mixed():
    s = '(--X a /\ ((p + q) = 5))'
    dvars, r, init, trans, win = past.translate(
        s, debug=True)
    assert init == '(~ a_prev1)', init
    r_ = '( a_prev1 /\ ( ( p + q ) = 5 ) )'
    assert r == r_, r


def test_parser_multi_previous():
    s = '--X --X a'
    tree = parser.parse(s)
    testers = dict()
    r = tree.flatten(testers=testers,
                     context='bool')
    assert r == '_aux1', r
    assert len(testers) == 2, testers


def test_context_checks():
    # boolean
    s = 'x + y'
    tree = parser.parse(s)
    tree.flatten(testers=dict,
                 context='arithmetic')
    with assert_raises(AssertionError):
        tree.flatten(testers=dict,
                     context='bool')
    # arithmetic
    s = 'x /\ y'
    tree = parser.parse(s)
    tree.flatten(testers=dict,
                 context='bool')
    with assert_raises(AssertionError):
        tree.flatten(testers=dict,
                     context='arithmetic')
    # nested
    s = 'y /\ (y + 3)'
    tree = parser.parse(s)
    with assert_raises(AssertionError):
        tree.flatten(testers=dict,
                     context='bool')


def test_isinstance_str():
    s = 's'
    r = stx.isinstance_str(s)
    assert r is True, r
    s = 0
    r = stx.isinstance_str(s)
    assert r is False, r


if __name__ == '__main__':
    test_context_checks()



"""Enumerate symbolically represented sets and relations.

Integers are decoded from binary to decimal.
"""
# Copyright 2015 by California Institute of Technology
# All rights reserved. Licensed under BSD-3.
#
import logging
import math

import natsort
import networkx as nx

from omega.logic import bitvector as bv
from omega.symbolic import bdd as scope
from omega.symbolic import symbolic


logger = logging.getLogger(__name__)


def dump_relation(
        u, aut,
        care_source=None,
        care_target=None,
        fname='edges.pdf'):
    """Dump graph of relation as PDF."""
    g = relation_to_graph(u, aut, care_source, care_target)
    h, umap = _format_nx(g)
    pd = nx.drawing.nx_pydot.to_pydot(h)
    pd.set_overlap(False)
    pd.write_pdf(fname, prog='neato')


def relation_to_graph(
        u, aut,
        care_source=None,
        care_target=None,
        care_bits=None):
    """Return enumerated graph from relation.

    @param u: relation as BDD node
    @param care_source: care set for source nodes
    @param care_target: care set for target nodes
    @type u, care_source, care_target: node in `aut.bdd`
    @rtype: `networkx.DiGraph`
    """
    assert u != aut.bdd.false
    t, care_relation = _make_table(u, aut, care_source,
                                   care_target, care_bits)
    c = _enumerate_bdd(u, aut.bdd, t,
                       care_relation, care_bits)
    # to nx graph
    g = nx.DiGraph()
    # fix an order of keys for lookup
    keys = list(aut.vars)
    umap = dict()
    for model in c:
        # model = model of relation = edge
        # source = model of set = node
        source = dict()
        target = dict()
        for var, value in model.items():
            if var in aut.vars:
                source[var] = value
            else:
                target[var] = value
        # map valuation to g node
        target = _unprime_any_primed(target)
        u = _find_or_add_model(source, umap, keys)
        v = _find_or_add_model(target, umap, keys)
        g.add_node(u, **source)
        g.add_node(v, **target)
        g.add_edge(u, v)
    return g


def print_nodes(
        u, dvars, bdd,
        care_set=None,
        care_bits=None):
    """Enumerate first-order models of a set.

    A set of nodes is defined over unprimed variables.

    @param dvars: table of unprimed variables
    @type bdd: `BDD`
    """
    assert scope.is_state_predicate(u), u.support
    if u == bdd.false:
        print('empty set')
        return
    if care_bits is not None:
        assert scope.support_issubset(u, care_bits, bdd), (
            support, care_bits)
    _print_enumeration(u, bdd, dvars, care_set, care_bits)


def print_edges(
        u, aut,
        care_set=None,
        care_bits=None):
    """Enumerate first-order models of a relation.

    A relation is defined over both primed and
    unprimed variables.

    @param `aut.vars`: table of unprimed variables
    @type `aut.bdd`: `BDD`
    """
    source = care_set
    target = care_set
    t, care_relation = _make_table(u, aut, source,
                                   target, care_bits)
    _print_enumeration(u, aut.bdd, t, care_relation,
                       care_bits)


def _print_enumeration(
        u, bdd, t,
        care_set,
        care_bits):
    """Print first-order models."""
    c = _enumerate_bdd(u, bdd, t, care_set, care_bits)
    r = list()
    keys = natsort.natsorted(t)
    for product in c:
        w = list()
        for x in keys:
            if x not in product:
                continue
            i = product[x]
            s = '({x} = {i})'.format(x=x, i=i)
            w.append(s)
        s = ' /\ '.join(w)
        r.append(s)
    e = '\n'.join(r)
    logger.debug((
        'enumeration of BDD node '
        '{u} with care set {c}\n').format(
            u=u, c=care_set))
    s = (
        'contains {n} expressions:\n\n{e}\n').format(
            e=e,
            n=len(c))
    print(s)


def _make_table(
        u, aut,
        care_source=None,
        care_target=None,
        care_bits=None):
    """Return symbol table with primed vars and care relation.

    The variables in `a.vars` should be unprimed.
    """
    bdd = aut.bdd
    care_relation = _care_relation(care_source, care_target,
                                   aut.prime, bdd)
    t = symbolic._prime_and_order_table(aut.vars)
    if care_bits is not None:
        assert scope.support_issubset(u, care_bits, bdd), (
            support, care_bits)
    return t, care_relation


def _care_relation(source, target, prime, bdd):
    """Return product `source` with primed `target`.

    @param source: care set for source nodes
    @param target: care set for target nodes
    @type source, target: BDD over unprimed variables
    """
    # Cartesian product of `care_set`
    if source is None or target is None:
        care_relation = None
        return care_relation
    primed_target = bdd.rename(target, prime)
    care_relation = source & primed_target
    return care_relation


def _enumerate_bdd(
        u, bdd, t,
        care_set=None,
        care_bits=None):
    """Enumerate first-order models of BDD `u`.

    @param care_set: enumerate only models in this set
    @param care_bits: enumerate over at least these bits
    """
    if u == bdd.false:
        return
    if care_set is not None:
        u &= care_set
        logging.debug((
            'enumerating BDD node {u}, '
            'with care set = {c}').format(
                u=u, c=care_set))
    c = list()
    for dbit in bdd.pick_iter(u, care_bits):
        dint = _bitfields_to_int_iter(dbit, t)
        c.extend(dint)
    return c


def _bitfields_to_int_iter(bits, t):
    """Enumerate set of integers (and Booleans).

    @param bits: partial assignment that maps bits to values
    @type bits: `dict`
    @param t: symbol table (with bitnames)
    @type t: `dict`
    """
    # any bits missing ?
    t_bits = set()
    for flatname, d in t.items():
        dt = d['type']
        if dt == 'bool':
            t_bits.add(flatname)
        elif dt == 'int':
            t_bits.update(d['bitnames'])
        else:
            raise Exception(
                'unknown type "{dt}"'.format(dt=dt))
    missing = set(bits).difference(t_bits)
    assert not missing, (
        'WARNING: missing bits:\n {b}\n'
        'from concrete table:\n {t}').format(
            b=missing, t=t)
    # init
    bits = dict(bits)
    model = dict()
    # bool first
    for flatname, d in t.items():
        if d['type'] != 'bool':
            continue
        if flatname in bits:
            model[flatname] = bits.pop(flatname)
    # integers
    sets = dict()
    for flatname, d in t.items():
        if d['type'] == 'bool':
            continue
        bitnames = d['bitnames']
        # missing ?
        if not set(bitnames).intersection(bits):
            continue
        # partial bitvector valuation
        bitvalues = list(map(bits.get, bitnames))
        bv._append_sign_bit(bitvalues, flatname, d)
        values = _enumerate_int(bitvalues)
        sets[flatname] = list(values)
    return _take_product_iter(sets, model)


def _enumerate_int(bitvalues, j=0):
    """Enumerate from a partial bitvector assignment.

    @param bitvalues: partial assignment to bits
    @type bitvalues: `list` of `str` or `None`
    @param j: index of current bit
    """
    n = len(bitvalues)
    assert j < n, (j, n)
    b = bitvalues[j]
    # sign ?
    if j == n - 1:
        if b is None:
            yield -2**j
            yield 0
        else:
            b = int(b)
            yield -b * 2**j
        return
    assert j < n - 1, (j, n)
    for v in _enumerate_int(bitvalues, j + 1):
        if b is None:
            yield v
            yield v + 2**j
        else:
            b = int(b)
            yield v + 2**j * b


def _take_product_iter(sets, model):
    """Iterate over all combinations of values in `sets`.

    The resulting combinations of values are
    used to complete the `model`.

    @param sets: map from variables to sets of values
    @type sets: `dict` of `set`
    @param model: partial assignment to Boolean variables
    @type model: `dict`
    """
    if not sets:
        yield model
        return
    var, values = sets.popitem()
    for m in _take_product_iter(sets, model):
        for v in values:
            m = dict(m)
            m[var] = v
            yield m


def _unprime_any_primed(model):
    """Trim any primed variables."""
    d = dict(model)
    suffix = "'"
    for k in d:
        if k.endswith(suffix):
            s = k[:-1]
            d[s] = d.pop(k)
    return d


def _find_or_add_model(model, umap, keys):
    """Return integer node for given valuation.

    If absent, then a fresh node is created.

    @type model: `dict`
    @type umap: `dict`
    """
    u = tuple(model[k] for k in keys)
    u = umap.setdefault(u, len(umap))
    return u


def _dump_graph_as_figure(g, fname):
    """Create a PDF file showing the graph `g`."""
    h, _ = _format_nx(g)
    pd = nx.drawing.nx_pydot.to_pydot(h)
    pd.write_pdf(fname)


def _format_nx(g, keys=None):
    """Return graph ready to be dumped.

    Nodes with same label over `keys` are identified.

    @type g: `networkx.DiGraph`
    @param keys: `list` of keys in node `dict` to show,
        in same order. By default all keys are shown.
    @rtype: `pydot.Graph`
    """
    h = nx.DiGraph()
    umap = dict()
    for u, d in g.nodes(data=True):
        if keys is None:
            keys = sorted(d)
        c = ['{var}={val}'.format(var=var, val=d[var])
             for var in keys if var in d]
        s = _square_conj(c)
        h.add_node(s)
        umap[u] = s
    for u, v in g.edges():
        us = umap[u]
        vs = umap[v]
        h.add_edge(us, vs)
    assert len(g) >= len(h), (g.nodes, h.nodes)
    assert len(g) == len(umap), (g.nodes, umap)
    return h, umap


def _square_conj(p, n=None, op=r'&and;'):
    """Return conjunction arranged vertically.

    The number of conjuncts per line is
    picked to make the result more balanced,
    closer to a square.
    The formatting characters are for `dot`.

    @param p: iterable of conjuncts
    @param n: number of items in `p`
        (useful when `p` is a generator)
    """
    if n is None:
        n = len(p)
    assert n > 0, n
    m = math.ceil(n**0.5)
    c = list()
    for i, s in enumerate(p):
        c.append(op)
        s = ' ({s}) '.format(s=s)
        c.append(s)
        if (i + 1) % m == 0:
            c.append('\l')
    # single line ?
    if n <= 2:
        c.pop(0)
    return ''.join(c)



"""How to compute from where Alice can reach her destination."""
from omega.logic import syntax as stx
from omega.symbolic import fixpoint as fx
from omega.symbolic import prime as prm
from omega.symbolic import temporal as trl


def do_it_yourself():
    """Compute states *from* where we can reach a goal."""
    aut, goal, action = spec()
    # Compute the least fixpoint,
    # so the states from where we can reach the `goal`
    # with one or more steps, or we are already there.
    vrs = aut.varlist['sys']
    primed_vrs = stx.prime_vars(vrs)
    r = aut.false
    rold = None
    # for sure != r, thus at least one iteration
    assert r != rold
    while r != rold:
        rold = r  # memo
        target = aut.replace_with_primed(vrs, r)
        pre = aut.exist(primed_vrs, action & target)
        r |= goal | pre
    assert prm.is_state_predicate(r)
    print_expr(r, aut)


def use_fixpoint_module():
    aut, goal, sys_action = spec()
    aut.varlist['env'] = list()  # controls all variables
    env_action = aut.true
    aut.plus_one = True
    aut.moore = True
    aut.build()  # store primed varlists
    r = fx.attractor(env_action, sys_action, goal, aut)
    print_expr(r, aut)


def spec():
    aut = trl.Automaton()
    aut.declare_variables(x=(1, 3), y=(0, 5), z='bool')
    aut.varlist['sys'] = ['x', 'y', 'z']
    goal = aut.to_bdd('(x = 2) /\ (y = 1) /\ z')
    action = aut.to_bdd('''
        (* type invariant *)
        /\ x \in 1..3
        /\ y \in 0..5
        /\ ((z <=> TRUE) \/ (z <=> FALSE))

        (* primed type invariant *)
        /\ x' \in 1..3
        /\ y' \in 0..5
        /\ ((z' <=> TRUE) \/ (z <=> FALSE))

        (* allowed changes *)
        /\ (x' = x + 1)
        /\ (y' = y - 2)
        /\ (z' <=> ~ z)
        ''')
    return aut, goal, action


def print_expr(u, aut):
    """BDD to expr in case of int and bool vars."""
    # currently, `to_expr` works with only
    # integer-valued variables
    aut.declare_constants(tmp=(0, 1))
    s = '\E z:  ({u} /\ (z <=> (tmp = 1)) )'.format(u=u)
    v = aut.to_bdd(s)
    s = aut.type_hint_for(['x', 'y', 'tmp'])
    care = aut.to_bdd(s)
    expr = aut.to_expr(v, care)
    print(expr)


if __name__ == '__main__':
    do_it_yourself()
    use_fixpoint_module()



"""A module for constructing behaviors of assembled state machines."""
# Copyright 2017 by California Institute of Technology
# All rights reserved. Licensed under 3-clause BSD.
#
from omega.games import enumeration as enum
from omega.logic import syntax as stx
from omega.symbolic import prime as prm


class History(object):
    """Record current and past values of variables.

    To initialize:

    ```python
    h = History(initial_state)
    ```

    or:

    ```python
    h.state = initial_state
    ```
    """

    def __init__(self):
        self.vars = dict()  # type hints
        self.state = dict()  # identifiers -> values
        # finite behavior
        self.past = list()  # list of earlier states

    def update(self, state):
        """Set `self.state` to `state`, update `self.past`.

        If you want to change the current state without
        appending the old state to `self.past`, then
        modify the attribute `self.state`.
        """
        self.past.append(self.state)
        self.state = state


class Assembly(History):
    """Register and step state machines.

    The assembly that results from these components should
    be a complete-system (aka closed-system, cf. Lamport).
    """

    def __init__(self):
        self.machines = dict()
        super(Assembly, self).__init__()
        self.state = None  # to call `init` first

    def init(self):
        """Return an initial state."""
        self.state = dict()
        for name, stm in self.machines.items():
            partial_state = self._init(stm, name)
            self._update_state(self.state, partial_state)

    def _init(self, machine, name):
        """Return partial state from `machine`.

        @param name: used to mangle hidden var names
        """
        partial = machine.init()
        return self._to_global_state(partial, name)

    def step(self):
        """Take a step and return the next state."""
        assert self.state is not None, 'first call method `init`'
        next_state = dict()
        for name, stm in self.machines.items():
            partial_state = self._step(stm, self.state, name)
            self._update_state(next_state, partial_state)
        # record history
        self.update(next_state)

    def _step(self, machine, state, name):
        """Return partial state update from `machine`.

        @param name: used to unmangle and mangle hidden
            var names
        """
        local = self._to_local_state(state, name, machine)
        partial = machine.step(local)
        return self._to_global_state(partial, name)

    def _to_local_state(self, global_state, name, machine):
        unmangled = omit_prefix(global_state, name)
        # ignore variables undeclared in `machine`
        local = {k: v for k, v in unmangled.items()
                 if k in machine.vars}
        return local

    def _to_global_state(self, local_state, name):
        vis = visible_vars(local_state)
        hid = hidden_vars(local_state)
        glob_hid = add_prefix(hid, name)
        _assert_disjoint(vis, glob_hid)
        vis.update(glob_hid)
        return vis

    def _update_state(self, state, partial):
        _assert_disjoint(partial, state)
        state.update(partial)


class AutomatonStepper(object):
    """Initialize and step a symbolic `Automaton`."""

    def __init__(self, aut):
        self.vars = aut.vars
        self.aut = aut
        self._init = aut.init['impl_sys']
        self._action = aut.action['impl']

    def init(self):
        """Return initial values of variables for this component."""
        r = self.aut.pick(self._init)
        return {k: v for k, v in r.items() if k in self.aut.varlist['sys']}

    def step(self, state):
        """Return next values of variables or `None`.

        @param state: `dict` that maps identifiers to values.

            If any unprimed identifiers in `support(action)` are
            missing from `state`, raise `ValueError`.

            If `action` is not enabled at `state`,
            then raise `ValueError`.

            For an `action` that depends on primed variables
            (Mealy machine), the `state` should include those.
        """
        action = self._action
        self._assert_support_assigned(action, state)
        u = self.aut.let(state, action)
        vrs = self.aut.support(u).union(self.aut.varlist["sys'"])
        primed_state = self.aut.pick(u, vrs)
        self._assert_unblocked(primed_state)
        return _unprime_state(primed_state)

    def _assert_support_assigned(self, action, state):
        unprimed = prm.unprimed_support(action, self.aut)
        missing = unprimed.difference(state)
        assert not missing, (missing, state)

    def _assert_unblocked(self, state):
        if state is None:
            raise ValueError((
                'action is not enabled '
                'at state: {state}').format(
                    state=state))


class EnumStrategyStepper(object):
    """Initialize and step an enumerated strategy."""

    def __init__(self, graph):
        assert graph.initial_nodes
        # self.vars = graph.vars
        self.graph = graph

    def init(self):
        """Return initial values for variables this component controls."""
        return self._pick_sys(self.graph.initial_nodes)

    def step(self, state):
        """Return next values for variables this component controls."""
        u = next(u for u in self.graph if self.graph.node[u] == state)
        return self._pick_sys(self.graph.successors(u))

    def _pick_sys(self, nodes):
        u = next(iter(nodes))
        d = self.graph.node[u]
        # TODO: consider defining `varlist` instead of
        # `inputs` and `outputs`
        return slice_dict(d, self.graph.outputs)


class Component(object):
    """Behave as a Moore state machine.

    When given "inputs", it updates its state
    by joining the given inputs to the "outputs" that
    it computes from the old state.

    Calling `step` you get the outputs from the current
    state, without changing the state. When using several
    `Component` instances, you should get their outputs
    by calling `step`, then feed them back as inputs by
    calling `update`.
    """

    def __init__(self, machine):
        self.machine = machine
        self.state = None  # `dict()` later

    def init(self):
        """Return initial values for variables that the component controls.

        Use this to get inputs to other machines that result
        after the first step of a behavior.
        """
        return self.machine.init()

    def step(self):
        """Return next values for variables component controls.

        Use this to get inputs to feed to other machines,
        from the second step onwards.
        """
        assert self.state is not None, 'first call method `update`'
        return self.machine.step(self.state)

    def update(self, inputs):
        """Update `self.state` based on `inputs` and current state.

        First call `init` from other machines,
        then feed the results to `update` here.
        Then repeatedly call `step` and `update`.
        """
        _inputs = set(self.machine.graph.inputs)
        assert set(inputs) == _inputs, (inputs, _inputs)
        # uninitialized yet ?
        if self.state is None:
            outputs = self.init()
        else:
            outputs = self.step()
        state = dict(inputs)
        state.update(outputs)
        self.state = state


class Scheduler(object):
    """A non-interleaving binary clock.

    Each step increments the variable "turn".
    For example:

    ```python
    sch = Scheduler(3)
    state_1 = dict(turn=1)

    state_2 = sch.step(state_1)
    >>> state_2
    {'turn': 2}

    state_3 = sch.step(state_2)
    >>> state_3
    {'turn': 0}
    ```
    """

    def __init__(self, n):
        dom = (0, n - 1)
        turn_type = dict(type='int', dom=dom)
        self.vars = dict(turn=turn_type)
        self._n = n

    def init(self):
        """Return initial values for the component's variable."""
        return dict(turn=0)

    def step(self, state):
        """Return assignment to variable "turn"."""
        next_turn = (state['turn'] + 1) % self._n
        return dict(turn=next_turn)


def visible_vars(vrs):
    """Slice dictionary `vrs`, omitting keys matching `_*`.

    @type vrs: `dict`
    @rtype: `dict`
    """
    return {k: v for k, v in vrs.items()
            if not k.startswith('_')}


def hidden_vars(vrs):
    """Slice dictionary `vrs`, keeping keys matching `_*`."""
    return {k: v for k, v in vrs.items()
            if k.startswith('_')}


def add_prefix(vrs, prefix):
    """Prepend `prefix` to hidden variables in `vrs`.

    @type vrs: `dict`
    @type prefix: `str`
    """
    d = dict()
    for k, v in vrs.items():
        if k.startswith('_'):
            name = prefix + k
            assert name not in d, (name, d)
        else:
            name = k
        d[name] = v
    return d


def omit_prefix(vrs, prefix):
    """Omit `prefix` from hidden variables in `vrs`.

    @type vrs: `dict`
    @type prefix: `str`
    """
    d = dict()
    for k, v in vrs.items():
        t = _omit_prefix(k, prefix)
        assert t not in d, (t, d)
        d[t] = v
    return d


def _omit_prefix(s, prefix):
    if s.startswith(prefix):
        return s.replace(prefix, '', 1)
    return s


def _unprime_state(primed_state):
    """Return same state but with identifiers unprimed."""
    return {stx.unprime(k): v for k, v in primed_state.items()}


def slice_dict(d, keys):
    """Return restriction of `d` to `keys`.

    @type d: `dict`
    @type keys: `set`
    """
    return {k: v for k, v in d.items()
            if k in keys}


def _assert_disjoint(a, b):
    """Raise `AssertionError` if common `dict` keys."""
    overlap = a.keys() & b.keys()
    assert not overlap, (overlap, a, b)


def enumerate_impl(spec):
    """Return enumerated controller."""
    # modifies env, sys (which serve as a reusable interface)
    spec.init['env'] = spec.init['impl_env']
    spec.init['sys'] = spec.init['impl_sys']
    spec.action['sys'] = spec.action['impl']
    graph = enum.action_to_steps(spec, qinit=spec.qinit)
    graph.inputs = spec.varlist['env']
    graph.outputs = spec.varlist['sys']
    # pprint.pprint(graph.nodes(data=True))
    return graph



"""Abstract syntax tree classes for logic."""
# Copyright 2014 by California Institute of Technology
# All rights reserved. Licensed under BSD-3.
#
import astutils


class Nodes(object):
    """Container of AST node classes for logic."""

    Terminal = astutils.Terminal

    class Var(astutils.Terminal):
        """Variable identifier."""

        def __init__(self, value, dtype='var'):
            super(Nodes.Var, self).__init__(value, dtype)

    class Bool(astutils.Terminal):
        """Boolean constant."""

        def __init__(self, value, dtype='bool'):
            super(Nodes.Bool, self).__init__(value, dtype)

    class Num(astutils.Terminal):
        """Numerical costant."""

        def __init__(self, value, dtype='num'):
            super(Nodes.Num, self).__init__(value, dtype)

    class Str(astutils.Terminal):
        """String constant."""

        def __init__(self, value, dtype='str'):
            # the parser removes quotes around `value`
            super(Nodes.Str, self).__init__(value, dtype)

    class Operator(astutils.Operator):
        def flatten(self, *arg, **kw):
            return ''.join([
                self.operator,
                '(',
                ', '.join(
                    x.flatten(*arg, **kw)
                    for x in self.operands),
                ')'])

    class Unary(astutils.Operator):
        """Unary operator."""

    class Binary(astutils.Operator):
        """Binary operator."""

        def flatten(self, *arg, **kw):
            # infix flattener for consistency with parser
            return ' '.join([
                '(',
                self.operands[0].flatten(*arg, **kw),
                self.operator,
                self.operands[1].flatten(*arg, **kw),
                ')'])

    class Comparator(Binary):
        """Binary predicate."""

    class Arithmetic(Binary):
        """Binary function."""



"""Utilities for creating formulas."""
# Copyright 2015 by California Institute of Technology
# All rights reserved. Licensed under BSD-3.
#
import logging
import math


logger = logging.getLogger(__name__)
PRIME = "'"
SUPPORTED_OPERATORS = {'&', '|', '/\\', '\/'}


def conj(iterable, sep='', op='/\\'):
    """Return binary conjunction in infix syntax."""
    return _associative_op(iterable, op, sep)


def disj(iterable, sep='', op=r'\/'):
    """Return binary disjunction in prefix syntax."""
    return _associative_op(iterable, op, sep)


def _associative_op(iterable, op, sep):
    """Apply associative binary operator `op`.

    @param sep: separator
    """
    if op not in SUPPORTED_OPERATORS:
        raise Exception(
            'operator "{op}" not supported.'.format(op=op))
    true = 'True'
    false = 'False'
    if op in ('&', '/\\'):
        true, false = false, true
    glue = (') ' + sep + op + ' (')
    # avoid consuming a generator
    h = [x for x in iterable if x]
    return _recurse_op(0, len(h), h, true, false, glue)


def _recurse_op(a, b, h, true, false, glue):
    """Apply binary operator recursively.

    @param a: start of sublist
    @type a: int in [0, len(h)]
    @param b: end of sublist
    @type b: int in [0, len(h)]
    @param h: `list`
    @param true, false: permutation of 'True', 'False'
    @param glue: used to concatenate a, b
    """
    n = b - a
    # empty ?
    if not n:
        return false
    # singleton ?
    if n == 1:
        return h[a]
    assert n > 1, n
    # power of 2 ?
    m = (n - 1).bit_length() - 1
    c = a + 2**m
    x = _recurse_op(a, c, h, true, false, glue)
    y = _recurse_op(c, b, h, true, false, glue)
    # controlling value ?
    # don't care ?
    if x == true or y == true:
        return true
    if x == false:
        return y
    if y == false:
        return x
    return '(' + x + glue + y + ')'


def paren(iterable):
    """Return generator that parenthesizes elements."""
    return ('(' + x + ')' for x in iterable)


def linear_disj(s, op='||'):
    """Return linear disjunction in prefix syntax."""
    b = ' {op} '.format(op=op)
    return b.join('({x})'.format(x=x) for x in s if x != '')


def linear_conj(s, op='&&'):
    """Return linear conjunction in prefix syntax."""
    b = ' {op} '.format(op=op)
    return b.join('({x})'.format(x=x) for x in s if x != '')


def disj_prefix(iterable, op='|', false='0', true='1'):
    """Return linear disjunction in prefix syntax."""
    return _prefix_linear(iterable, op, false, true)


def conj_prefix(iterable, op='&', false='0', true='1'):
    """Return linear conjunction in prefix syntax."""
    false, true = true, false
    return _prefix_linear(iterable, op, false, true)


# TODO: recursive version
def _prefix_linear(s, op, false, true):
    """Apply associative binary operator linearly.

    @param s: container
    @param op: operator
    @param false, true: values if treating `op` as disjunction
    """
    if not s:
        return false
    u = s[0]
    for v in s[1:]:
        # controlling value ?
        if u == true:
            break
        if v == true:
            u = true
            break
        u = op + ' ' + u + ' ' + v
    return u


def conj_intersection(s, r, paren=True):
    if paren:
        return conj(
            '({x})'.format(x=x) for x in s if x in r)
    else:
        return conj(
            '{x}'.format(x=x) for x in s if x in r)


def conj_neg(s, paren=True):
    if paren:
        return conj('!({x})'.format(x=x) for x in s)
    else:
        return conj('!{x}'.format(x=x) for x in s)


def recurse_binary_log_space(f, x, n):
    """Apply associative binary operator `f` to generator `x`."""
    logger.info('++ recurse binary')
    assert n > 0, n
    if n == 1:
        return next(x)
    assert n > 1, n
    m = (n - 1).bit_length() - 1
    c = 2**m
    assert c < n <= 2 * c, (n, c)
    a = recurse_binary_log_space(f, x, c)
    b = recurse_binary_log_space(f, x, n - c)
    logger.info('-- done binary {n} items'.format(n=n))
    return f(a, b)


def recurse_binary(f, x, bdd=None):
    """Recursively traverse binary tree of computation."""
    logger.info('++ recurse binary')
    n = len(x)
    logger.debug('{n} items left to recurse'.format(n=n))
    assert n > 0, n
    if n == 1:
        assert len(x) == 1, x
        return x.pop()
    assert n > 1, n
    # largest power of 2 less than n
    m = (n - 1).bit_length() - 1
    c = 2**m
    assert c < n <= 2 * c, (n, c)
    left = x[:c]
    right = x[c:]
    del x[:]
    a = recurse_binary(f, left, bdd)
    b = recurse_binary(f, right, bdd)
    logger.info(bdd)
    logger.info('-- done recurse binary ({n} items)'.format(n=n))
    return f(a, b)


def _compute_as_binary_tree(f, x):
    """Return result of applying operator `f`.

    In-place computation.
    Non-recursive implementation.
    """
    logger.debug('++ start binary tree')
    assert len(x) > 0
    while len(x) > 1:
        n = len(x)
        logger.debug('Binary at: {n}'.format(n=n))
        k = int(math.floor(0.5 * n))
        # consume the power of 2
        for i in range(k):
            j = 2 * i
            a = x[j]
            b = x[j + 1]
            x[i] = f(a, b)
        # has last element ?
        if n % 2 == 1:
            x[k] = x[2 * k]
            # empty tail
            del x[k + 1:]
        else:
            del x[k:]
        n_ = len(x)
        assert n_ == n - k, (n_, n - k)
    assert len(x) == 1, len(x)
    logger.debug('-- done binary tree')
    return x[0]


def _compute_as_binary_tree_simple(f, x):
    """Return result of applying operator `f`.

    Delete level d only after computing level (d + 1).
    Non-recursive implementation.
    """
    logger.debug('++ start binary tree')
    assert len(x) > 0
    while len(x) > 1:
        n = len(x)
        # halve
        r = [f(a, b) for a, b in zip(x[::2], x[1::2])]
        # has last element ?
        if n % 2 == 1:
            r.append(x[-1])
        # empty tail
        x = r
        # assert
        n_ = len(x)
        k = int(math.floor(0.5 * n))
        assert n_ == n - k, (n_, n - k)
    assert len(x) == 1, len(x)
    logger.debug('-- done binary tree')
    return x.pop()


def _linear_operator(f, x):
    """Return result of applying linearly operator `f`."""
    logger.debug('++ start linear operator')
    assert len(x) > 0
    n = len(x)
    for i in range(1, n):
        x[0] = f(x[0], x.pop())
    assert len(x) == 1, len(x)
    logger.debug('-- done linear operator')
    return x.pop()


def _linear_operator_simple(f, x):
    """Return result of applying linearly operator `f`."""
    logger.debug('++ start simple linear operator')
    assert len(x) > 0
    u = x[0]
    for v in x[1:]:
        u = f(u, v)
    logger.debug('-- done simple linear operator')
    return u


def test_binary_operators():
    for n in range(1, 1500):
        a = range(n)
        f = _plus
        x0 = _compute_as_binary_tree(f, list(a))
        x1 = _compute_as_binary_tree_simple(f, list(a))
        x2 = _linear_operator(f, list(a))
        x3 = _linear_operator_simple(f, list(a))
        x4 = recurse_binary(f, list(a))
        z = sum(a)
        assert x0 == z, (x0, z)
        assert x1 == z, (x1, z)
        assert x2 == z, (x2, z)
        assert x3 == z, (x3, z)
        assert x4 == z, (x4, z)
        print(z)


def _plus(x, y):
    return x + y


def prime_vars(vrs):
    """Return `list` of primed variables from `vrs`."""
    return [prime(var) for var in vrs]


def unprime_vars(vrs):
    """Return `list` of primed variables from `vrs`."""
    return [unprime(var) for var in vrs]


def prime(var):
    """Return primed variable."""
    assert not isprimed(var), var
    return var + PRIME


def unprime(var):
    """Return unprimed variable."""
    assert isprimed(var), var
    s = var[:-1]
    assert not isprimed(s), s
    return s


def isprimed(var):
    """Return `True` iff `var` ends with prime."""
    return var[-1] == PRIME


def isinstance_str(s):
    """Raise `AssertionError` if `s` is not a "string"."""
    try:
        s + 's'
    except TypeError:
        return False
    return True



"""Solving a game to win a Streett(1) specification."""
import logging

import networkx as nx
from omega.symbolic import logicizer
from omega.symbolic import enumeration as sym_enum
from omega.games import gr1
from omega.games import enumeration as enum
from omega.automata import TransitionSystem


log = logging.getLogger('omega.games.enumeration')
log.addHandler(logging.StreamHandler())
log.setLevel(logging.DEBUG)


def semi_symbolic():
    """Example using a semi-enumerated state machine.

    Instructive variants:

    - `formula = "x'"`
    - `self_loops = True`
    - `aut.moore = False`
    """
    g = TransitionSystem()
    g.owner = 'sys'
    g.vars = dict(x='bool')
    g.env_vars.add('x')
    g.add_path(range(11))
    g.add_edge(10, 10)
    g.add_edge(10, 0, formula="x")
    # symbolic
    aut = logicizer.graph_to_logic(
        g, 'nd', ignore_initial=True, self_loops=False)
    aut.init['env'] = 'nd = 1'
    aut.win['<>[]'] = aut.bdds_from(' ~ x')
    aut.win['[]<>'] = aut.bdds_from('nd = 0')
    aut.qinit = '\A \A'
    aut.moore = True
    aut.plus_one = True
    print(aut)
    # compile to BDD
    z, yij, xijk = gr1.solve_streett_game(aut)
    gr1.make_streett_transducer(z, yij, xijk, aut)
    # print t.bdd.to_expr(t.action['sys'][0])
    r = aut.action['sys']
    # aut.bdd.dump('bdd.pdf', roots=[r])
    g = enumerate_controller(aut)
    h, _ = sym_enum._format_nx(g)
    pd = nx.drawing.nx_pydot.to_pydot(h)
    pd.write_pdf('game_states.pdf')
    print('Enumerated strategy has {n} nodes.'.format(
        n=len(g)))
    print(('Winning set:', aut.bdd.to_expr(z)))
    print('{n} BDD nodes in total'.format(
        n=len(aut.bdd)))


def enumerate_controller(aut):
    aut.init['env'] = aut.init['impl_env']
    aut.init['sys'] = aut.init['impl_sys']
    aut.action['sys'] = aut.action['impl']
    g = enum.action_to_steps(aut, qinit=aut.qinit)
    return g


if __name__ == '__main__':
    semi_symbolic()



"""Slugsin translator to BDDs that uses iteration.

This module translates Boolean formulas in prefix format
to binary decision diagrams.


Reference
=========

Slugsin syntax:
    https://github.com/VerifiableRobotics/
    slugs/blob/master/doc/input_formats.md
"""
# Copyright 2016 by California Institute of Technology
# All rights reserved. Licensed under BSD-3.
#
from omega.symbolic.bdd import Lexer


OPERATORS = {'!', '&', '|', '^', '\A', '\E', '\S'}
BINARY_OPERATORS = {'&', '|', '^', '\A', '\E', '\S'}
BINARY = {'AND', 'OR', 'XOR', 'FORALL', 'EXISTS', 'RENAME'}


class Parser(object):
    """Parser for prefix syntax with buffers.

    Avoids recursion by using a stack.
    """
    # Context-sensitive grammar, so cannot use PLY

    def __init__(self):
        self.lexer = Lexer()
        self.tokens = self.lexer.tokens
        self._binary = {'AND', 'OR', 'XOR'}

    def parse(self, data, bdd):
        self.lexer.lexer.input(data)
        mem = None
        r = self._increase(mem, bdd)
        # empty stack ?
        tok = self.lexer.lexer.token()
        if not tok:
            return r
        # error
        s = tok.value
        while True:
            tok = self.lexer.lexer.token()
            if not tok:
                break
            s += tok.value
        raise Exception(
            'syntax error: remaining characters: {s}'.format(
                s=s))

    def _increase(self, mem, bdd):
        stack = list()
        need = 1
        while need > 0:
            need = self._push(stack, mem, need, bdd)
        return self._reduce(stack, bdd)

    def _push(self, stack, mem, need, bdd):
        tok = self.lexer.lexer.token()
        if not tok:
            raise Exception('syntax error: stream ended early')
        t = tok.type
        if t == 'NAME':
            r = bdd.var(tok.value)
            need -= 1
        elif t == 'NUMBER':
            u = int(tok.value)
            if u == 0:
                r = bdd.false
            elif u == 1:
                r = bdd.true
            else:
                r = bdd._add_int(u)
            need -= 1
        elif t == 'NOT':
            r = tok.value
        elif t in BINARY:
            r = tok.value
            need += 1
        elif t == 'QUESTION':
            tok = self.lexer.lexer.token()
            i = int(tok.value)
            assert 0 <= i < len(mem), (i, mem)
            r = mem[i]
            need -= 1
        elif t == 'DOLLAR':
            tok = self.lexer.lexer.token()
            n = int(tok.value)
            mem = list()
            for i in range(n):
                s = self._increase(mem, bdd)
                mem.append(s)
            r = mem[-1]
            need -= 1
        else:
            raise Exception(
                'unknown token type "{t}"'.format(t=t))
        stack.append(r)
        return need

    def _reduce(self, stack, bdd):
        while len(stack) > 1:
            for i, t in enumerate(reversed(stack)):
                if not isinstance(t, str):
                    continue
                if t in OPERATORS:
                    break
            assert t in OPERATORS, (
                'unknown operator "{t}"'.format(t=t))
            operator = t
            k = len(stack) - i
            if operator == '!':
                n = 1
            elif t in BINARY_OPERATORS:
                n = 2
            else:
                raise Exception(
                    'unknown operator "{t}"'.format(t=t))
            operands = stack[k:k + n]
            for i in range(n):
                stack.pop(k)
            stack.pop(k - 1)
            # print(operator)
            # print(operands)
            u = bdd.apply(operator, *operands)
            stack.insert(k - 1, u)
        (r,) = stack
        return r


parser = Parser()


def add_expr(e, bdd):
    """Add to `bdd` a node for Boolean expression `e`.

    @param e: expression in SLUGSIN syntax
    @type e: `str`
    @param t: symbol table
    @type t: `dict`
    @type bdd: `BDD`

    @return: node in `bdd` corresponding to `e`
    @rtype: `int`
    """
    u = parser.parse(e, bdd)
    return u



#!/usr/bin/env python
"""Test `omega.symbolic.fol`."""
import logging
import pprint

from nose import tools as nt
from omega.symbolic import fol as _fol


log = logging.getLogger('dd')
log.setLevel(logging.WARNING)
log = logging.getLogger('omega.logic')
log.setLevel(logging.WARNING)


def test_declare():
    # Boolean-valued variable
    fol = _fol.Context()
    bdd = fol.bdd
    fol.declare(x='bool')
    assert 'x' in fol.vars, fol.vars
    dx = fol.vars['x']
    assert 'type' in dx, dx
    type_x = dx['type']
    assert type_x == 'bool', type_x
    assert 'x' in bdd.vars
    assert 'x_0' not in bdd.vars
    # integer-valued variable
    fol = _fol.Context()
    bdd = fol.bdd
    fol.declare(y=(0, 2))
    assert 'y' in fol.vars, fol.vars
    dy = fol.vars['y']
    assert 'type' in dy, dy
    type_dy = dy['type']
    assert type_dy == 'int', type_dy
    # regression regarding bit naming convention
    assert 'y_0' in bdd.vars
    assert 'y_1' in bdd.vars
    assert 'y_2' not in bdd.vars
    assert 'y' not in bdd.vars
    # primed vars
    fol = _fol.Context()
    bdd = fol.bdd
    d = {"y'": (0, 1)}
    fol.declare(**d)
    assert "y'" in fol.vars, fol.vars
    assert "y_0'" in bdd.vars, bdd.vars
    assert "y'_0" not in bdd.vars, bdd.vars
    # adding same vars twice
    fol = _fol.Context()
    fol.declare(x='bool')
    fol.declare(x='bool')
    # mismatch with existing var
    with nt.assert_raises(ValueError):
        fol.declare(x=(1, 5))
    # mixed new and existing
    fol.declare(x='bool', y=(0, 5))
    with nt.assert_raises(ValueError):
        fol.declare(x='bool', y=(3, 15))


def test_support():
    fol = _fol.Context()
    bdd = fol.bdd
    s = 'True'
    u = fol.add_expr(s)
    r = fol.support(u)
    assert r == set(), r
    # single variable
    fol.declare(x=(0, 10), y='bool')
    with nt.assert_raises(AssertionError):
        u = fol.add_expr('x')
    with nt.assert_raises(AssertionError):
        u = fol.add_expr(5)
    s = 'x = 1'
    u = fol.add_expr(s)
    r = fol.support(u)
    assert r == {'x'}, r
    s = '(x < 2) \/ ~ y'
    u = fol.add_expr(s)
    r = fol.support(u)
    assert r == {'x', 'y'}, r


def test_replace():
    fol = _fol.Context()
    fol.declare(
        x=(0, 12), y=(0, 12),
        z=(0, 3), w='bool')
    # with vars
    u = fol.add_expr('x = 1')
    subs = {'x': 'y'}
    u = fol.let(subs, u)
    u_ = fol.add_expr('y = 1')
    assert u == u_, (u, u_)
    u = fol.add_expr('x = 1')
    # with mismatched substitutions
    subs = {'x': 'z'}
    with nt.assert_raises(AssertionError):
        fol.let(subs, u)
    subs = {'x': 'w'}
    with nt.assert_raises(AssertionError):
        fol.let(subs, u)
    # with constants
    # Boolean
    u = fol.add_expr('w')
    subs = {'w': True}
    u = fol.let(subs, u)
    assert u == fol.bdd.true
    u = fol.add_expr('w')
    subs = {'w': False}
    u = fol.let(subs, u)
    assert u == fol.bdd.false
    # integer
    u = fol.add_expr('z = 2')
    subs = {'z': 2}
    u = fol.let(subs, u)
    assert u == fol.bdd.true
    u = fol.add_expr('z = 2')
    subs = {'z': 3}
    u = fol.let(subs, u)
    assert u == fol.bdd.false
    # outside coarse type bound
    u = fol.add_expr('z = 2')
    subs = {'z': 4}
    with nt.assert_raises(AssertionError):
        fol.let(subs, u)
    # no effect
    u = fol.add_expr('w')
    subs = {'z': True}
    v = fol.let(subs, u)
    assert u == v


# requires `dd.cudd`
def replace_with_bdd():
    fol = _fol.Context()
    fol.declare(x='bool', y='bool')
    u = fol.add_expr('x => y')
    subs = {'x': fol.bdd.true}
    u = fol.replace_with_bdd(u, subs)
    u_ = fol.add_expr('y')
    assert u == u_, (u, u_)


def test_quantifiers():
    """Test syntax for rigid quantification."""
    fol = _fol.Context()
    bdd = fol.bdd
    fol.declare(x=(0, 5))
    # no `qvars`
    u = fol.add_expr('TRUE')
    r = fol.exist(set(), u)
    assert r == u, (r, u)
    u = fol.add_expr('FALSE')
    r = fol.exist(set(), u)
    assert r == u, (r, u)
    u = fol.add_expr('x = 5')
    r = fol.exist(set(), u)
    assert r == u, (r, u)
    # TYPENESS !!!
    s = '\E x: (x = 7)'
    u = fol.add_expr(s)
    assert u == bdd.true
    u = fol.add_expr('x = 3')
    u = fol.forall(['x'], u)
    assert u == bdd.false
    # fine-grained type hint ignored
    u = fol.add_expr('x <= 5')
    u = fol.forall(['x'], u)
    assert u == bdd.false
    # coarse type-hint `(-1 + 2**a)..(-1 + 2**b)` effective
    u = fol.add_expr('x < 8')
    u = fol.forall(['x'], u)
    assert u == bdd.true
    # \E
    u = fol.add_expr('x = 2')
    u = fol.exist(['x'], u)
    assert u == bdd.true
    u = fol.add_expr('x < 0')
    u = fol.exist(['x'], u)
    assert u == bdd.false
    u = fol.add_expr('x > 5')
    u = fol.exist(['x'], u)
    assert u == bdd.true
    u = fol.add_expr('(6 <= x) /\ (x < 8)')
    u = fol.exist(['x'], u)
    assert u == bdd.true
    u = fol.add_expr('x > 7')
    u = fol.exist(['x'], u)
    assert u == bdd.false


def test_pick():
    fol = _fol.Context()
    fol.declare(x='bool', y=(0, 2))
    u = fol.add_expr('x')
    p = fol.pick(u, care_vars=['x'])
    for i in range(10):
        q = fol.pick(u, care_vars=['x'])
        assert p == q, (p, q)
    u = fol.add_expr('False')
    p = fol.pick(u, care_vars=['x'])
    assert p is None, p


def test_pick_iter():
    fol = _fol.Context()
    fol.declare(x='bool', y=(0, 2))
    u = fol.add_expr('True')
    gen = fol.pick_iter(u, care_vars=['x'])
    r = list(gen)
    assert len(r) == 2, r
    u = fol.add_expr('False')
    gen = fol.pick_iter(u, care_vars=['x'])
    r = list(gen)
    assert len(r) == 0, r
    u = fol.add_expr('x')
    gen = fol.pick_iter(u, care_vars=['x'])
    r = list(gen)
    assert len(r) == 1, r
    u = fol.add_expr('x')
    gen = fol.pick_iter(u, care_vars=['x', 'y'])
    r = list(gen)
    assert len(r) == 4, r


def test_define():
    c = _fol.Context()
    c.declare(x=(9, 35))
    c.declare(y=(1, 5))
    c.declare(z=(-3, 10))
    e = '''
        a == x + y > 3
        b == z - x <= 0
        c == a /\ b
        '''
    c.define(e)
    # pprint.pprint(c.op)
    # a
    s = c.op['a']
    s_ = '( ( x + y ) > 3 )'
    assert s == s_, s
    u = c.op_bdd['a']
    u_ = c.add_expr(s_)
    assert u == u_, u
    # b
    s = c.op['b']
    s_ = '( ( z - x ) <= 0 )'
    assert s == s_, s
    u = c.op_bdd['b']
    u_ = c.add_expr(s_)
    assert u == u_, u
    # c
    s = c.op['c']
    s_ = '( a /\ b )'
    assert s == s_, s
    u = c.op_bdd['c']
    # caution:  operator names *have* been substituted in
    # the BDD, unlike the expression in `c.op['c']` above.
    s_ = '( ( ( x + y ) > 3 )  /\  ( ( z - x ) <= 0 ) )'
    u_ = c.add_expr(s_)
    assert u == u_, u


def test_add_expr():
    fol = _fol.Context()
    bdd = fol.bdd
    u = fol.add_expr('False')
    assert u == bdd.false, bdd.to_expr(u)
    fol.declare(x=(0, 100), y=(5, 23))
    u = fol.add_expr('x < y + 5')
    v = fol.add_expr('x - 5 < y')
    assert u == v, (u, v)


def test_to_expr():
    fol = _fol.Context()
    fol.bdd.configure(reordering=True)
    fol.declare(x=(-14, 100), y=(5, 23))
    u = fol.add_expr('(1 <= x) /\ (x <= 3)')
    s = fol.to_expr(u, show_limits=True)
    s_ = ' /\\ x \\in -128 .. 127\n /\\ (x \in 1 .. 3)'
    assert s == s_, (s, s_)
    care = fol.add_expr('(-14 <= x)  /\  (x <= 100)')
    s = fol.to_expr(u, care=care, show_dom=True)
    s_ = (
        ' /\\ x \\in -14 .. 100\n'
        ' /\\ (x \in 1 .. 3)\n'
        ' /\\ care expression')
    assert s == s_, (s, s_)


def test_apply():
    fol = _fol.Context()
    bdd = fol.bdd
    fol.declare(x='bool')
    u = fol.add_expr('x')
    v = fol.add_expr(' ~ x')
    w = fol.apply('and', u, v)
    assert w == bdd.false
    w = fol.apply('or', u, v)
    assert w == bdd.true


def test_reorder():
    fol = _fol.Context()
    bdd = fol.bdd
    fol.declare(x=(0, 7), y=(0, 3))
    top_bit = bdd.var_at_level(0)
    if top_bit.startswith('x'):
        var = 'y'
    elif top_bit.startswith('y'):
        var = 'x'
    else:
        raise AssertionError(str(top_bit))
    shift_vars = {var: dict(level=0)}
    _fol.reorder(shift_vars, fol)
    bit_0 = bdd.var_at_level(0)
    bit_1 = bdd.var_at_level(1)
    assert bit_0.startswith(var), bit_0
    assert bit_1.startswith(var), bit_1


if __name__ == '__main__':
    test_quantifiers()



"""Utilities for manipulating primes (x') and support."""
# Copyright 2015-2017 by California Institute of Technology
# All rights reserved. Licensed under BSD-3.
#
from omega.logic import syntax as stx
from omega.symbolic import bdd as sym_bdd


# functions in transition to here
is_state_predicate = sym_bdd.is_state_predicate
is_proper_action = sym_bdd.is_proper_action
prime = sym_bdd.prime
unprime = sym_bdd.unprime


def is_primed_state_predicate(u, fol):
    """Return `True` if `u` does not depend on unprimed variables.

    Only constant parameters (rigid variables) can appear
    unprimed in `u`. Any flexible variables in `u` should
    be primed.

    An identifier that is declared only unprimed is assumed
    to be a rigid variable. If a primed sibling is declared,
    then the identifier is assumed to be a flexible variable.
    """
    unprimed_vars = (is_variable(k, fol) for k in unprimed_support(u, fol))
    return not any(unprimed_vars)


def is_action_of_player(action, player, aut):
    """Return `True` if `action` constrains only `player`.

    The `player` is represented by the variables in
    `aut.varlist[player]`.
    """
    primed = primed_support(action, aut)
    vrs = aut.vars_of_players([player])
    vrs_p = stx.prime_vars(vrs)
    r = primed.issubset(vrs_p)
    return r


def rigid_support(u, fol):
    """Return constants that `u` depends on."""
    unprimed = unprimed_support(u, fol)
    return {k for k in unprimed if is_constant(k, fol)}


def flexible_support(u, fol):
    """Return unprimed variables that `u` depends on."""
    unprimed = unprimed_support(u, fol)
    return {k for k in unprimed if is_variable(k, fol)}


def split_support(u, fol):
    """Return unprimed, primed identifiers `u` depends on.

    This function exists as an optimization over calling
    separately `unprimed_support` and `primed_support`.
    This function calls `fol.support` once, as opposed to
    twice when performing two separate calls.

    If optimization is irrelevant, then call those other
    functions, because readability counts [PEP 20].
    """
    support = fol.support(u)
    primed = {k for k in support if stx.isprimed(k)}
    unprimed = support - primed
    return unprimed, primed


def unprimed_support(u, fol):
    """Return unprimed identifiers that `u` depends on."""
    return {k for k in fol.support(u) if not stx.isprimed(k)}


def primed_support(u, fol):
    """Return primed identifiers that `u` depends on.

    These identifiers include both variables and constants.
    """
    return {k for k in fol.support(u) if stx.isprimed(k)}


def prime(u, fol):
    """Prime state predicate `u`."""
    support = fol.support(u)
    # all identifiers are unprimed ?
    assert not any(stx.isprimed(name) for name in support), support
    # avoid priming constants
    # (no primed identifiers are declared for constants)
    vrs = {name for name in support if is_variable(name, fol)}
    let = {var: stx.prime(var) for var in vrs}
    return fol.let(let, u)


def is_constant(name, fol):
    """Return `True` if `name` is declared as constant."""
    return not is_variable(name, fol)


def is_variable(name, fol):
    """Return `True` if `name` is declared as variable.

    The identifier `name` should be unprimed.
    """
    return stx.prime(name) in fol.vars


def pairwise_disjoint(c):
    """Return whether elements in `c` are pairwise disjoint."""
    union = set().union(*c)
    n = sum(len(u) for u in c)
    return n == len(union)


def pick(c):
    """Return an element from container `c`.

    If `c` is empty, return `None`.
    """
    return next(iter(c), None)



"""Algorithms for generalized Streett and Rabin games.

References
==========

Roderick Bloem, Barbara Jobstmann, Nir Piterman,
Amir Pnueli, Yaniv Sa'ar
    "Synthesis of reactive(1) designs"
    Journal of Computer and System Sciences
    Vol.78, No.3, pp.911--938, 2012


Robert Konighofer
    "Debugging formal specifications with
    simplified counterstrategies"
    Master's thesis
    Inst. for Applied Information Processing and Communications,
    Graz University of Technology, 2009
"""
# Copyright 2016 by California Institute of Technology
# All rights reserved. Licensed under BSD-3.
#
import logging
import copy

from omega.symbolic import bdd as sym_bdd
from omega.symbolic.bdd import is_state_predicate
from omega.symbolic import fixpoint as fx
from omega.symbolic import fol as _fol
from omega.symbolic import symbolic
from omega.symbolic import temporal as trl


logger = logging.getLogger(__name__)


def solve_streett_game(aut, rank=1):
    r"""Return winning set and iterants for Streett(1) game.

    The returned value takes into account actions and
    liveness, not initial conditions (i.e., it is the
    fixpoint before reasoning about initial conditions).

    Expects "env" and "sys" players. Writes:

      - `aut.varlist["env'"]`
      - `aut.varlist["sys'"]`

    @param aut: compiled game with <>[] \/ []<> winning
    @type aut: `temporal.Automaton`
    """
    assert rank == 1, 'only rank 1 supported for now'
    assert aut.bdd.vars or not aut.vars, (
        'first call `Automaton.build`')
    assert len(aut.win['<>[]']) > 0
    assert len(aut.win['[]<>']) > 0
    env_action = aut.action['env']
    sys_action = aut.action['sys']
    aut.build()
    z = aut.true
    zold = None
    while z != zold:
        zold = z
        cox_z = fx.step(env_action, sys_action, z, aut)
        xijk = list()
        yij = list()
        for goal in aut.win['[]<>']:
            goal &= cox_z
            y, yj, xjk = _attractor_under_assumptions(goal, aut)
            z &= y
            xijk.append(xjk)
            yij.append(yj)
    assert is_state_predicate(z), z.support
    return z, yij, xijk


def _attractor_under_assumptions(goal, aut):
    """Targeting `goal`, under unconditional assumptions."""
    env_action = aut.action['env']
    sys_action = aut.action['sys']
    xjk = list()
    yj = list()
    y = aut.false
    yold = None
    while y != yold:
        yold = y
        cox_y = fx.step(env_action, sys_action, y, aut)
        unless = cox_y | goal
        xk = list()
        for safe in aut.win['<>[]']:
            x = fx.trap(env_action, sys_action,
                        safe, aut, unless=unless)
            xk.append(x)
            y |= x
        yj.append(y)
        xjk.append(xk)
    return y, yj, xjk


def make_streett_transducer(z, yij, xijk, aut):
    """Return I/O `temporal.Automaton` implementing strategy.

    An auxiliary variable `_goal` is added,
    to represent the counter of recurrence goals.
    """
    winning = z
    assert is_realizable(winning, aut)
    _warn_moore_mealy(aut)
    # declare goal counter var
    n_goals = len(aut.win['[]<>'])
    c = '_goal'
    c_max = n_goals - 1
    vrs = {c: (0, c_max)}
    aut.declare_variables(**vrs)
    aut.varlist['sys'].append(c)
    # compile transducer with refined shared BDD
    env_init = aut.init['env']
    sys_init = aut.init['sys']
    env_action = aut.action['env']
    sys_action = aut.action['sys']
    holds = aut.win['<>[]']
    goals = aut.win['[]<>']
    # compute strategy from iterates
    # \rho_1: switch goals
    rho_1 = aut.false
    for i, goal in enumerate(goals):
        ip = (i + 1) % len(goals)
        s = "({c} = {i}) /\ ({c}' = {ip})".format(c=c, i=i, ip=ip)
        u = aut.add_expr(s)
        u &= goal
        rho_1 |= u
    zstar = _controllable_action(z, aut)
    rho_1 &= zstar
    # \rho_2: descent in basin
    rho_2 = aut.false
    for i, yj in enumerate(yij):
        s = "({c} = {i}) /\ ({c}' = {i})".format(c=c, i=i)
        count = aut.add_expr(s)
        rho_2j = aut.false
        basin = yj[0]
        for y in yj[1:]:
            # steps leading to next basin
            ystar = _controllable_action(basin, aut)
            rim = y & ~ basin
            u = rim & ystar
            rho_2j |= u
            basin |= y
        u = rho_2j & count
        rho_2 |= u
    # \rho_3: persistence holds
    rho_3 = aut.false
    for i, xjk in enumerate(xijk):
        s = "({c} = {i}) /\ ({c}' = {i})".format(c=c, i=i)
        count = aut.add_expr(s)
        rho_3j = aut.false
        used = aut.false
        for xk in xjk:
            assert len(xk) == len(holds), xk
            for x, hold in zip(xk, holds):
                # steps leading to next wait
                xstar = _controllable_action(x, aut)
                stay = x & ~ used
                used |= x
                u = stay & xstar
                u &= hold
                rho_3j |= u
        u = rho_3j & count
        rho_3 |= u
    # \rho
    u = rho_1 | rho_2
    u |= rho_3
    # counter `c` limits
    s = '{c} \in 0..{n}'.format(c=c, n=c_max)
    u &= aut.add_expr(s)
    # `sys_action` is already in the `\rho`
    # next is "useful" only if `env_action` depends on `y'`
    if not aut.plus_one:
        u |= ~ env_action
        if aut.moore:
            u = aut.forall(aut.varlist["env'"], u)
    assert u != aut.false
    aut.action['impl'] = u
    symbolic._assert_support_moore(u, aut)
    # initial condition for counter
    # (no closure taken for counter)
    s = '{c} = 0'.format(c=c)
    count = aut.add_expr(s)
    _make_init(count, winning, aut)
    return aut.action['impl']


def solve_rabin_game(aut, rank=1):
    """Return winning set and iterants for Rabin(1) game.

    @param aut: compiled game with <>[] & []<> winning
    @type aut: `temporal.Automaton`
    """
    assert rank == 1, 'only rank 1 supported for now'
    assert aut.bdd.vars or not aut.vars, (
        'first call `Automaton.build`')
    # TODO: can these assertions be removed elegantly ?
    assert len(aut.win['<>[]']) > 0
    assert len(aut.win['[]<>']) > 0
    aut.build()
    z = aut.false
    zold = None
    zk = list()
    yki = list()
    xkijr = list()
    while z != zold:
        zold = z
        xijr = list()
        yi = list()
        for hold in aut.win['<>[]']:
            y, xjr = _cycle_inside(zold, hold, aut)
            z |= y
            xijr.append(xjr)
            yi.append(y)
        zk.append(z)
        yki.append(yi)
        xkijr.append(xijr)
    assert is_state_predicate(z), z.support
    return zk, yki, xkijr


def _cycle_inside(z, hold, aut):
    """Cycling through goals, while staying in `hold`."""
    env_action = aut.action['env']
    sys_action = aut.action['sys']
    cox_z = fx.step(env_action, sys_action, z, aut)
    g = cox_z | hold
    y = aut.true
    yold = None
    while y != yold:
        yold = y
        cox_y = fx.step(env_action, sys_action, y, aut)
        inside = cox_y & g
        xjr = list()
        for goal in aut.win['[]<>']:
            x, xr = _attractor_inside(inside, goal, aut)
            xjr.append(xr)
            y &= x
    return y, xjr


def _attractor_inside(inside, goal, aut):
    env_action = aut.action['env']
    sys_action = aut.action['sys']
    xr = list()
    x = aut.false
    xold = None
    while x != xold:
        xold = x
        cox_x = fx.step(env_action, sys_action, x, aut)
        x = cox_x | goal
        x &= inside
        x |= xold
        xr.append(x)
    return x, xr


def make_rabin_transducer(zk, yki, xkijr, aut):
    """Return O/I transducer for Rabin(1) game."""
    winning = zk[-1]
    assert is_realizable(winning, aut)
    _warn_moore_mealy(aut)
    dvars = dict(aut.vars)
    n_holds = len(aut.win['<>[]'])
    n_goals = len(aut.win['[]<>'])
    # add transducer memory as two indices:
    # - `w`: persistence hold index
    # - `c`: recurrence goal index
    w = '_hold'
    c = '_goal'
    n_w = n_holds - 1 + 1  # last value used as "none"
    n_c = n_goals - 1
    vrs = {w: (0, n_w), c: (0, n_c)}
    aut.declare_variables(**vrs)
    aut.varlist['sys'].extend([w, c])
    # compile
    env_init = aut.init['env']
    sys_init = aut.init['sys']
    env_action = aut.action['env']
    sys_action = aut.action['sys']
    goals = aut.win['[]<>']
    # compute strategy from iterates
    # \rho_1: descent in persistence basin
    s = "({c}' = {c}) /\ ({w}' = {none})".format(
        c=c, w=w, none=n_holds)
    count = aut.add_expr(s)
    rho_1 = aut.false
    basin = zk[0]
    for z in zk[1:]:
        zstar = _controllable_action(basin, aut)
        rim = z & ~ basin
        u = rim & zstar
        u &= count
        rho_1 |= u
        basin = z
    rho_2 = aut.false
    rho_3 = aut.false
    rho_4 = aut.false
    basin = aut.false
    for z, yi, xijr in zip(zk, yki, xkijr):
        cox_basin = fx.step(env_action, sys_action,
                            basin, aut)
        rim = z & ~ basin
        rim &= ~ cox_basin
        # rho_2: pick persistence set
        s = "({c}' = {c}) /\ ({w} = {none})".format(
            c=c, w=w, none=n_holds)
        count = aut.add_expr(s)
        u = rim & count
        v = aut.false
        for i, y in enumerate(yi):
            s = "{w}' = {i}".format(w=w, i=i)
            count = aut.add_expr(s)
            ystar = _controllable_action(y, aut)
            q = count & ystar
            v |= q
        u &= v
        rho_2 |= u
        # rho_3: descent in recurrence basin
        s = (
            "({c}' = {c}) /\ "
            "({w} /= {none}) /\ "
            "({w}' = {w})").format(
                c=c, w=w, none=n_holds)
        count = aut.add_expr(s)
        u = rim & count
        v = aut.false
        for i, xjr in enumerate(xijr):
            for j, (xr, goal) in enumerate(zip(xjr, goals)):
                s = (
                    "({c} = {j}) /\ "
                    " ({w} = {i})").format(c=c, w=w, i=i, j=j)
                count = aut.add_expr(s)
                x_basin = xr[0]
                p = aut.false
                for x in xr[1:]:
                    xstar = _controllable_action(x_basin, aut)
                    q = xstar & ~ x_basin
                    q &= x
                    p |= q
                    x_basin = x
                p &= count
                p &= ~ goal
                v |= p
        u &= v
        rho_3 |= u
        # rho_4: advance to next recurrence goal
        u = aut.false
        for j, goal in enumerate(goals):
            jp = (j + 1) % len(goals)
            s = "({c} = {j}) /\ ({c}' = {jp})".format(
                c=c, j=j, jp=jp)
            count = aut.add_expr(s)
            p = count & goal
            u |= p
        s = (
            "({w} /= {none}) /\ "
            "({w}' = {w})").format(
                c=c, w=w, none=n_holds)
        count = aut.add_expr(s)
        u &= count
        u &= rim
        v = aut.false
        for i, y in enumerate(yi):
            s = "{w} = {i}".format(w=w, i=i)
            count = aut.add_expr(s)
            ystar = _controllable_action(y, aut)
            q = count & ystar
            v |= q
        u &= v
        rho_4 |= u
        # update
        basin = z
    # \rho
    u = rho_1 | rho_2
    u |= rho_3
    u |= rho_4
    # counter limits
    s = '({w} \in 0..{n_w}) /\ ({c} \in 0..{n_c})'.format(
        w=w, n_w=n_w, c=c, n_c=n_c)
    u &= aut.add_expr(s)
    if not aut.plus_one:
        u |= ~ env_action
        if aut.moore:
            u = aut.forall(aut.varlist["env'"], u)
    assert u != aut.false
    aut.action['impl'] = u
    symbolic._assert_support_moore(u, aut)
    # initial condition for counter
    s = '({c} = 0) /\ ({w} = {none})'.format(
        c=c, w=w, none=n_holds)
    count = aut.add_expr(s)
    _make_init(count, winning, aut)
    return aut.action['impl']


def is_realizable(z, aut):
    """Return `True` if, and only if, `aut` wins from `z`.

    @param z: bdd node
    @param type: compiled `temporal.Automaton`
    """
    env_init = aut.init['env']
    sys_init = aut.init['sys']
    sys_action = aut.action['sys']
    evars = aut.varlist['sys']
    uvars = aut.varlist['env']
    # common errors
    assert env_init != aut.false, 'vacuous spec'
    # realizable ?
    qinit = aut.qinit
    if qinit == '\A \A':
        w = ~ env_init | sys_init
        if w != aut.true:
            print(
                'WARNING: `qinit = "\A \A"` but '
                'not `EnvInit => SysInit`')
        w = aut.exist(aut.varlist["sys'"], sys_action)
        w |= ~ env_init
        if w != aut.true:
            print(
                "WARNING: `qinit = '\A \A'` but "
                "not `EnvInit => (\E y':  SysNext)`")
        # \A env, sys:
        #    env_init => /\ sys_init
        #                /\ z
        u = sys_init & z
        u |= ~ env_init
        r = (u == aut.true)
        msg = (
            'some initial states are losing:\n'
            '`\A x, y:  EnvInit => (SysInit /\ Win)`')
    elif qinit == '\E \E':
        # \E env, sys:  /\ env_init
        #               /\ sys_init
        #               /\ z
        u = sys_init & z
        u &= env_init
        u = aut.exist(evars + uvars, u)
        r = (u == aut.true)
        msg = (
            'no state satisfies:\n'
            '`EnvInit /\ SysInit /\ Win`')
    elif qinit == '\A \E':
        assert not aut.moore
        # \A env:  \E sys:
        #    (\E sys:  env_init) => /\ env_init
        #                           /\ sys_init
        #                           /\ z
        a = aut.exist(evars, env_init)
        u = sys_init & z
        u &= env_init
        u |= ~ a
        u = aut.exist(evars, u)
        u = aut.forall(uvars, u)
        r = (u == aut.true)
        msg = (
            'cannot for each x pick a y:\n'
            '`\A x:  \E y:\n'
            '    (\E y:  EnvInit) => /\ EnvInit\n'
            '                        /\ SysInit\n'
            '                        /\ Win`')
    elif qinit == '\E \A':
        # \E sys:  \A env:
        #    (\E sys:  env_init) => /\ env_init
        #                           /\ sys_init
        #                           /\ z
        a = aut.exist(evars, env_init)
        u = sys_init & z
        u &= env_init
        u |= ~ a
        u = aut.forall(uvars, u)
        u = aut.exist(evars, u)
        r = (u == aut.true)
        msg = (
            'cannot pick y that works for all x:\n'
            '`\E y:  \A x:\n'
            '    (\E y:  EnvInit) => /\ EnvInit\n'
            '                        /\ SysInit\n'
            '                        /\ Win`')
    else:
        raise ValueError(
            'unknown `qinit` value "{qinit}"'.format(
                qinit=qinit))
    if not r:
        print(msg)
    return r


def _controllable_action(target, aut):
    """Return controllable transitions for progress.

    Compared to CPre, this has "half" the quantification.
    """
    env_action = aut.action['env']
    sys_action = aut.action['sys']
    u = sym_bdd.prime(target, aut)
    if aut.plus_one:
        # sys_action /\ (env_action => target')
        u |= ~ env_action
        u &= sys_action
    else:
        # env_action => (sys_action /\ target')
        u &= sys_action
        u |= ~ env_action
    if aut.moore:
        # \A uvars'
        u = aut.forall(aut.varlist["env'"], u)
    return u


def _make_init(internal_init, win, aut):
    """Return initial conditions for implementation.

    Depending on `aut.qinit`,
    synthesize the initial condition `aut.env_init`
    using the winning states `win`.

    @param internal_init: initial condition for
        internal variables.
    """
    qinit = aut.qinit
    # impl qinit = '\A \A'  # we synthesize `env_init` below
    env_init = aut.init['env']
    new_sys_init = aut.init['sys'] & internal_init & win
    # synthesize initial predicate
    if qinit in ('\A \A', '\A \E', '\E \E'):
        new_env_init = env_init & new_sys_init
    elif qinit == '\E \A':
        env_bound = aut.exist(aut.varlist['sys'], env_init)
        u = env_init & new_sys_init
        u |= ~ env_bound
        sys_bound = aut.forall(aut.varlist['env'], u)
        new_env_init = env_bound & sys_bound
    else:
        raise ValueError(
            'unknown `qinit` value "{qinit}"'.format(
                qinit=qinit))
    assert new_env_init != aut.false
    assert new_sys_init != aut.false
    assert sym_bdd.is_state_predicate(new_env_init)
    assert sym_bdd.is_state_predicate(new_sys_init)
    aut.init['impl_env'] = new_env_init
    aut.init['impl_sys'] = new_sys_init


def _warn_moore_mealy(aut):
    """Warn the user if they define actions suspect of error."""
    env_action = aut.action['env']
    sys_action = aut.action['sys']
    moore = aut.moore
    env_support = aut.support(env_action)
    sys_support = aut.support(sys_action)
    env_depends_on_primed_sys = env_support.intersection(
        aut.varlist["sys'"])
    sys_depends_on_primed_env = sys_support.intersection(
        aut.varlist["env'"])
    r = True
    if moore and sys_depends_on_primed_env:
        r = False
        print(
            'WARNING: Moore sys, but sys depends on '
            'primed env vars:\n {r}'.format(
                r=sys_depends_on_primed_env))
    if not moore and env_depends_on_primed_sys:
        r = False
        print((
            'WARNING: Mealy sys, and assumption depends on '
            'primed sys vars.\n'
            'If env has to be Mealy too, '
            'then you can get cyclic dependencies.\n'
            'Primed sys vars:\n {r}').format(
                r=env_depends_on_primed_sys))
    if env_depends_on_primed_sys:
        r = False
        print((
            'ATTENTION: assumption depends on primed sys vars:\n'
            '{r}\n'
            'Is a Mealy env realistic for your problem ?').format(
                r=env_depends_on_primed_sys))
    return r


def trivial_winning_set(aut_streett):
    """Return set of trivially winning nodes for Streett(1).

    @return: `(trivial, aut_streett)` where:
        - `trivial`: node in `aut_streett.bdd`
        - `aut_streett`: `temporal.Automaton`
    """
    aut_rabin = trl.default_rabin_automaton()
    aut_rabin.bdd = aut_streett.bdd
    aut_rabin.vars = copy.deepcopy(aut_streett.vars)
    # swap vars and actions
    aut_rabin.varlist['env'] = list(aut_streett.varlist['sys'])
    aut_rabin.varlist['sys'] = list(aut_streett.varlist['env'])
    aut_rabin.action['env'] = aut_streett.action['sys']
    aut_rabin.action['sys'] = aut_streett.action['env']
    aut_rabin.init['env'] = aut_streett.init['sys']
    aut_rabin.init['sys'] = aut_streett.init['env']
    win = [~ w for w in aut_streett.win['<>[]']]
    aut_rabin.win['[]<>'] = win
    # needed because we changed bdd
    aut_rabin.win['<>[]'] = [aut_rabin.true]
    # solve
    win_streett, _, _ = solve_streett_game(aut_streett)
    zk, _, _ = solve_rabin_game(aut_rabin)
    win_rabin = zk[-1]
    # find trivial win set
    # win_rabin_ = _copy_bdd(win_rabin,
    #                        aut_rabin.bdd, aut_streett.bdd)
    trivial = win_streett & ~ win_rabin
    return trivial, aut_streett


def _map_nested_lists(f, x, *arg, **kw):
    """Recursively map lists, with non-lists at the bottom.

    Useful for applying `dd.bdd.copy_bdd` to several lists.
    """
    if isinstance(x, list):
        return [_map_nested_lists(f, y, *arg, **kw) for y in x]
    else:
        return f(x, *arg, **kw)


def _copy_bdd(u, a, b):
    """Copy bdd `u` from manager `a` to `b`.

    No effect if `a is b`.
    """
    if a is b:
        return u
    return a.copy(u, b)
